'''
Created on 23 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import cos, isnan, linspace, math, pi, prod, sin, sqrt
from shapely.geometry import LineString, Point, Polygon


class EllipseLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def getEllipseContour(centre, semiXAxis, semiYAxis, azim, npoints=40):
        x0, y0 = centre.x, centre.y
        if isnan(azim):
            # CIRCLE
            result = Polygon([Point((x0 + semiXAxis * cos(t), y0 + semiXAxis * sin(t))) for t in linspace(0, 2 * pi, npoints)])
        else:
            if (semiXAxis < semiYAxis):
                azim = (azim + pi / 2.0) if (azim < 0) else (azim - pi / 2.0)
            result = Polygon([EllipseLib.parametricEquation(x0, y0, semiXAxis, semiYAxis, azim, t) for t in linspace(0, 2 * pi, npoints)])
        # result[-1] = result[0]
        return result

    @staticmethod
    def getEllipseMainAxis(centre, semiXAxis, semiYAxis, azim):
        if isnan(azim):
            return None

        semiAxis = semiXAxis if (semiXAxis >= semiYAxis) else semiYAxis
        p1 = Point((centre.x - semiAxis * cos(azim), centre.y - semiAxis * sin(azim)))
        p2 = Point((centre.x + semiAxis * cos(azim), centre.y + semiAxis * sin(azim)))
        return LineString((p1, p2))

    @staticmethod
    def getEllipsePerimeter(semiXAxis, semiYAxis, order=None):
        return EllipseLib.__getEllipsePerimeter_ramanujan(semiXAxis, semiYAxis)
        # return EllipseLib.__getEllipsePerimeter_infinite_series(semiXAxis, semiYAxis)

    @staticmethod
    def __getEllipsePerimeter_ramanujan(semiXAxis, semiYAxis):
        # https://en.wikipedia.org/wiki/Ellipse
        # According to Ramanujan 2nd formula
        diff = semiXAxis - semiYAxis
        summ = semiXAxis + semiYAxis
        if (0.0 == summ):
            return 0.0
        h = diff * diff / (summ * summ)
        perim = pi * summ * (1.0 + (3.0 * h) / (10.0 + sqrt(4.0 - 3.0 * h)))
        return float(perim)

    @staticmethod
    def __doubleFactorial(n):
        # https://en.wikipedia.org/wiki/Double_factorial
        # 9!! = 9 * 7 * 5 * 3 * 1 = 945
        # 8!! = 8 * 6 * 4 * 2 = 384
        return prod(list(range(n, 1, -2)))

    @staticmethod
    def __getEllipsePerimeter_infinite_series(semiXAxis, semiYAxis, order=100):
        diff = semiXAxis - semiYAxis
        summ = semiXAxis + semiYAxis
        if (0.0 == summ):
            return 0.0
        h = diff * diff / (summ * summ)

        acc = 1.0 + 0.25 * h
        for n in range(2, order):
            tmp = EllipseLib.__doubleFactorial(2 * n - 3) / ((2 ** n) * math.factorial(n))
            acc += tmp * tmp * h ** n
        return float(pi * summ * acc)

    @staticmethod
    def parametricEquation(x0, y0, semiXAxis, semiYAxis, theta, t):
        x = x0 + semiXAxis * cos(theta) * cos(t) - semiYAxis * sin(theta) * sin(t)
        y = y0 + semiXAxis * sin(theta) * cos(t) + semiYAxis * cos(theta) * sin(t)
        return ((x, y))



'''
Created on 28 sept. 2020

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arctan2, cos, mean, sin, sum


class SVFLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def svf1981(heights, widths):
        '''
        Based on:
        Oke, T. R. 1981. "Canyon geometry and the nocturnal heat island:Comparison of scale model and
        field observations." Journal of Climatology, 1 (237-254).

        Cited in:
        Swaid, Hanna. 1993. "The Role of Radiative-Convective Interaction in Creating the Microclimate 
        of Urban Street Canyons." Boundary-Layer Meteorology 64 (3): 231-59. doi: 10.1007/BF00708965.
        '''
        n = len(widths)
        tmp = [cos(arctan2(heights[i], widths[i])) for i in range(n)]
        return float(mean(tmp))

    @staticmethod
    def svf2018(heights, widths):
        '''
        Based on:
        Bernard, J., Bocher, E., Petit, G. and Palominos, S. (2018) "Sky View Factor Calculation in 
        Urban Context: Computational Performance and Accuracy Analysis of Two Open and Free GIS Tools.",
        Climate, 6(3), p. 60. doi: 10.3390/cli6030060.

        Modified by:
        Rodler, A., and Leduc, T. (2019). "Local climate zone approach on local and micro scales: Dividing 
        the urban open space.", Urban Climate, 28(March), 100457. doi: 10.1016/j.uclim.2019.100457
        '''
        n = len(widths)
        tmp = [sin(arctan2(heights[i], widths[i])) for i in range(n)]
        return float(1.0 - sum(tmp) / n)



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from t4gpd.comfort.algo.CommonsLib import CommonsLib
from t4gpd.comfort.algo.ConstantsLib import ConstantsLib


class TmrtOutLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def assess_tmrt_out(N, SR01Up_1, SR01Dn_1, IR01UpCo_1, IR01DnCo_1):
        SDIF, Idn = CommonsLib.SDIF_Idn(N, SR01Up_1)

        #=======================================================
        # Calculation of OUT_SET based on Watanabe et al. (2014)
 
        # net direct solar radiation to the body [W.m-2]
        dir_body = ConstantsLib.ah * ConstantsLib.fp * Idn
        # net diffuse solar radiation to the body [W.m-2]
        diff_body = (ConstantsLib.ah * ConstantsLib.alpha_eff 
                     * ((SDIF + SR01Dn_1) / 2))
        # net longwave to the body [W.m-2]
        longw_body = (ConstantsLib.alpha_l * ConstantsLib.alpha_eff
                      * ((IR01UpCo_1 + IR01DnCo_1) / 2))

        # OUT_MRT after Watanabe etal. (2014)
        Tmrt_OUT = ((dir_body + diff_body + longw_body)
                    / (ConstantsLib.alpha_l * ConstantsLib.alpha_eff
                       * ConstantsLib.sigma_B)) ** 0.25 - ConstantsLib.T_kel

        return Tmrt_OUT



'''
Created on 25 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.ops import polygonize, triangulate, unary_union
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, MultiPoint
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.Epsilon import Epsilon
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STVoronoiPartition(object):
    '''
    classdocs
    '''

    def __init__(self, points, crs=None):
        '''
        Constructor
        '''
        if isinstance(points, MultiPoint):
            _points = points
            self.crs = crs
        elif isinstance(points, GeoDataFrame):
            _points = []
            for geom in points.geometry:
                _points += GeomLib.getListOfShapelyPoints(geom)
            _points = MultiPoint(_points)
            self.crs = points.crs if crs is None else crs
        else:
            raise IllegalArgumentTypeException(points, 'MultiPoint or GeoDataFrame')

        self.tin = triangulate(_points)

    def __append(self, myDict, a, b, circumCircleCenter):
        if a in myDict:
            if b in myDict[a]:
                myDict[a][b].append(circumCircleCenter)
            else:
                myDict[a][b] = [circumCircleCenter]
        else:
            myDict[a] = dict({ b:[circumCircleCenter] })
        
    def __getCircumCircleCenter(self, a, b, c):
        if GeomLib.areAligned(a, b, c):
            ab = GeomLib.distFromTo(a, b)
            bc = GeomLib.distFromTo(b, c)
            ca = GeomLib.distFromTo(c, a)
            if Epsilon.EPSILON > min([ab, bc, ca]):
                return a
            diam = max([ab, bc, ca])
            if diam == ab:
                return GeomLib.getMidPoint(a, b) 
            elif diam == bc:
                return GeomLib.getMidPoint(b, c) 
            elif diam == ca:
                return GeomLib.getMidPoint(c, a)

        line1 = GeomLib.getLineSegmentBisector(a, b)
        line2 = GeomLib.getLineSegmentBisector(b, c)
        return GeomLib.intersect_line_line(line1, line2)

    def getDelaunayTriangles(self):
        return GeoDataFrame([{'geometry':t} for t in self.tin], crs=self.crs)

    def run(self):
        myDict = dict()
        for t in self.tin:
            a, b, c = sorted(t.exterior.coords[:-1])
            circumCircleCenter = self.__getCircumCircleCenter(a, b, c)

            self.__append(myDict, a, b, circumCircleCenter)
            self.__append(myDict, b, c, circumCircleCenter)
            self.__append(myDict, a, c, circumCircleCenter)

        edges = []
        for a in myDict.keys():
            for b in myDict[a].keys():
                if (2 == len(myDict[a][b])):
                    edges.append(LineString(myDict[a][b]))
        edgesUnion = unary_union(edges)

        # Contour network polygonization
        patches = polygonize(edgesUnion)

        # rows = [{ 'geometry':edges }]
        rows = [{ 'gid':gid, 'geometry':patch } for gid, patch in enumerate(patches)]
        return GeoDataFrame(rows, crs=self.crs)



'''
Created on 16 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import MultiLineString, Point
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.NeighborhoodLib import NeighborhoodLib


class STRoadNeighborhood(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, roads, fromPoints, maxDists):
        '''
        Constructor
        '''
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, 'GeoDataFrame')
        if not isinstance(fromPoints, (GeoDataFrame, Point)):
            raise IllegalArgumentTypeException(fromPoints, 'GeoDataFrame or single Point')
        if not isinstance(maxDists, (str, float, int)):
            raise IllegalArgumentTypeException(maxDists, 'String fieldname, float or int values')

        self.roads = roads

        if isinstance(fromPoints, GeoDataFrame):
            if maxDists in fromPoints:
                self.fromPointsAndMaxDists = list(zip(fromPoints.centroid, fromPoints[maxDists]))
            elif isinstance(maxDists, (float, int)):
                self.fromPointsAndMaxDists = list(zip(fromPoints.centroid, [maxDists] * len(fromPoints)))
        elif isinstance(maxDists, (float, int)):
            self.fromPointsAndMaxDists = [fromPoints, maxDists]
        else:
            raise Exception('%s is neither a float value nor a fromPoints fieldname!' % maxDists)

    def run(self):
        rows = []
        for gid, (fromPoint, maxDist) in enumerate(self.fromPointsAndMaxDists):
            gdf = NeighborhoodLib.neighborhood(self.roads, fromPoint, maxDist)
            rows.append({
                'gid':gid, 'fromPoint': fromPoint.wkt, 'maxDist': maxDist,
                'geometry': MultiLineString([geom for geom in gdf.geometry])
                })
            gid += 1
        return GeoDataFrame(rows, crs=self.roads.crs)



'''
Created on 23 juin 2022

self.author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arccos, pi, sqrt
from shapely.geometry import Polygon
from t4gpd.commons.GeomLib3D import GeomLib3D


class Triangle3D(object):
    '''
    classdocs
    '''

    def __init__(self, ptA, ptB, ptC):
        '''
        Constructor
        '''
        self.ptA, self.ptB, self.ptC = ptA, ptB, ptC  # A REVOIR

    def centroid(self):
        x = (self.ptA[0] + self.ptB[0] + self.ptC[0]) / 3.0
        y = (self.ptA[1] + self.ptB[1] + self.ptC[1]) / 3.0
        z = (self.ptA[2] + self.ptB[2] + self.ptC[2]) / 3.0
        return [ x, y, z ]

    def divideBy4OnSphere(self):
        midAB = self.__midOfEdgeOnSphere(self.ptA, self.ptB)
        midBC = self.__midOfEdgeOnSphere(self.ptB, self.ptC)
        midCA = self.__midOfEdgeOnSphere(self.ptC, self.ptA)
        return [
            Triangle3D(self.ptA, midAB, midCA),
            Triangle3D(midAB, self.ptB, midBC),
            Triangle3D(midCA, midBC, self.ptC),
            Triangle3D(midBC, midCA, midAB)
            ]

    def __midOfEdgeOnSphere(self, pt1, pt2):
        x = (pt1[0] + pt2[0]) / 2.0
        y = (pt1[1] + pt2[1]) / 2.0
        z = (pt1[2] + pt2[2]) / 2.0
        invNorm = 1.0 / sqrt(x ** 2 + y ** 2 + z ** 2)
        return [ x * invNorm, y * invNorm, z * invNorm ]

    def area3D(self):
        # Heron's formula
        len1 = GeomLib3D.distFromTo(self.ptA, self.ptB)
        len2 = GeomLib3D.distFromTo(self.ptB, self.ptC)
        len3 = GeomLib3D.distFromTo(self.ptC, self.ptA)
        sp = (len1 + len2 + len3) / 2.0
        return sqrt(sp * (sp - len1) * (sp - len2) * (sp - len3))

    def sphericalArea(self, mockupCentre=[0, 0, 0], radius=1.0):
        # Girard's theorem - area of the spherical triangle
        oa = GeomLib3D.vector_to(mockupCentre, self.ptA)
        ob = GeomLib3D.vector_to(mockupCentre, self.ptB)
        oc = GeomLib3D.vector_to(mockupCentre, self.ptC)

        normalOAB = GeomLib3D.unitVector(GeomLib3D.crossProduct(oa, GeomLib3D.vector_to(self.ptA, self.ptB)))
        normalOBC = GeomLib3D.unitVector(GeomLib3D.crossProduct(ob, GeomLib3D.vector_to(self.ptB, self.ptC)))
        normalOCA = GeomLib3D.unitVector(GeomLib3D.crossProduct(oc, GeomLib3D.vector_to(self.ptC, self.ptA)))

        alpha = arccos(GeomLib3D.dotProduct(
            GeomLib3D.unitVector(GeomLib3D.crossProduct(oa, normalOAB)),
            GeomLib3D.unitVector(GeomLib3D.crossProduct(oa, normalOCA))
            ))
        beta = arccos(GeomLib3D.dotProduct(
            GeomLib3D.unitVector(GeomLib3D.crossProduct(ob, normalOAB)),
            GeomLib3D.unitVector(GeomLib3D.crossProduct(ob, normalOBC))
            ))
        gamma = arccos(GeomLib3D.dotProduct(
            GeomLib3D.unitVector(GeomLib3D.crossProduct(oc, normalOBC)),
            GeomLib3D.unitVector(GeomLib3D.crossProduct(oc, normalOCA))
            ))

        return (alpha + beta + gamma - pi) * radius ** 2

    def toPolygon(self):
        return Polygon([self.ptA, self.ptB, self.ptC, self.ptA])



'''
Created on 18 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, MultiLineString, Point

from t4gpd.commons.BoundingBox import BoundingBox
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class AngularAbscissa(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, vpoint_x='vpoint_x', vpoint_y='vpoint_y', nRays=64):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        if vpoint_x not in inputGdf:
            raise Exception('Unknown vpoint_x field name: %s!' % (vpoint_x))
        if vpoint_y not in inputGdf:
            raise Exception('Unknown vpoint_y field name: %s!' % (vpoint_y))

        self.vpoint_x, self.vpoint_y = vpoint_x, vpoint_y
        self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)

    def runWithArgs(self, row):
        # Use a buffer to avoid bugs
        geom = row.geometry.buffer(0.01, -1)
        bbox = BoundingBox(geom)
        maxRayLen = max(bbox.width(), bbox.height())

        vp = Point((float(row[self.vpoint_x]), float(row[self.vpoint_y])))
        rays = []
        for sdCos, sdSin in self.shootingDirs:
            rp = Point((vp.x + sdCos * maxRayLen, vp.y + sdSin * maxRayLen))
            ray = LineString((vp, rp))
            tmp = geom.intersection(ray)
            # QUID DE:
            # from shapely.ops import nearest_points 
            # _, nearestNode = nearest_points(vp, tmp)
            nodes = GeomLib.getListOfShapelyPoints(tmp)
            nearestNode, minDist = None, float('inf')
            for node in nodes:
                dist = vp.distance(node)
                if (0 < dist) and (dist < minDist):
                    minDist = dist
                    nearestNode = node
            if nearestNode is not None:
                rays.append(LineString((vp, nearestNode)))

        return { 'geometry':MultiLineString(rays) }



'''
Created on 25 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime

from t4gpd.commons.AngleLib import AngleLib

import matplotlib.pyplot as plt
from numpy  import arange
from t4gpd.energy.Dogniaux import Dogniaux
from t4gpd.energy.Perez import Perez
from t4gpd.energy.Perraudeau import Perraudeau
from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut


class PlotDirectNormalIrradiance(object):
    '''
    classdocs
    '''

    @staticmethod
    def plot(dt=datetime(2020, 3, 21), ofile=None):
        # PEREZ MODEL PARAMETERS:
        delta = 0.12
        epsilon = 6.3
        dayInYear = dt.timetuple().tm_yday

        xDeg = arange(0, 91, 1)
        xRad = AngleLib.toRadians(xDeg)

        y1 = [Dogniaux.directNormalIrradiance(x) for x in xRad]
        y2 = [Perez.directSolarIrradiance(x, dayInYear, delta, epsilon) for x in xRad]
        y3 = [Perraudeau.directNormalIrradiance(x) for x in xRad]
        y4 = [PerrinDeBrichambaut.directNormalIrradiance(x, skyType=PerrinDeBrichambaut.PURE_SKY) for x in xRad]
        y5 = [PerrinDeBrichambaut.directNormalIrradiance(x, skyType=PerrinDeBrichambaut.STANDARD_SKY) for x in xRad]
        y6 = [PerrinDeBrichambaut.directNormalIrradiance(x, skyType=PerrinDeBrichambaut.POLLUTED_SKY) for x in xRad]
        y7 = [Perez.diffuseSolarIrradiance(x, dayInYear, delta) for x in xRad]
        y8 = [PerrinDeBrichambaut.diffuseSolarIrradiance(x, skyType=PerrinDeBrichambaut.PURE_SKY) for x in xRad]
        y9 = [PerrinDeBrichambaut.diffuseSolarIrradiance(x, skyType=PerrinDeBrichambaut.STANDARD_SKY) for x in xRad]
        ya = [PerrinDeBrichambaut.diffuseSolarIrradiance(x, skyType=PerrinDeBrichambaut.POLLUTED_SKY) for x in xRad]

        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.5 * 8.26))
        ax.set_title('Direct Normal Irradiance (DNI) + Diffuse Solar Irradiance', fontsize=16)
        ax.set_xlabel('Solar Altitude Angle', fontsize=16)
        ax.set_ylabel('Solar Irradiance [W.m$^{-2}$]', fontsize=16)
        ax.set_ylim(0, 1200)
        ax.grid(True)
        _lines = ax.plot(xDeg, y1, 'b-', xDeg, y2, 'r-', xDeg, y3, 'k-', xDeg, y4, 'm-.',
                         xDeg, y5, 'g-.', xDeg, y6, 'c-.', xDeg, y7, 'r--', xDeg, y8, 'm--',
                         xDeg, y9, 'g--', xDeg, ya, 'c--')
        ax.legend(_lines, ["Dogniaux's DNI", "Perez's DNI", "Perraudeau's DNI",
                           "Perrin de Br.'s DNI (Pure sky)",
                           "Perrin de Br.'s DNI (Standard sky)",
                           "Perrin de Br.'s DNI (Polluted sky)",
                           "Perez's diffuse model",
                           "Perrin de Br.'s diffuse model (Pure sky)",
                           "Perrin de Br.'s diffuse model (Standard sky)",
                           "Perrin de Br.'s diffuse model (Polluted sky)"
                           ], loc='upper center', ncol=2, fontsize=12)
        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches='tight')
        plt.close(fig)


# PlotDirectNormalIrradiance.plot(ofile='/tmp/irrad.pdf')



'''
Created on 18 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import pi, sqrt
from t4gpd.commons.ChrystalAlgorithm import ChrystalAlgorithm
from t4gpd.commons.DiameterLib import DiameterLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class CircularityIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    @staticmethod
    def runWithArgs(row):
        geom = row.geometry
        area, perim = geom.area, geom.length
        _, diamLen, _ = DiameterLib.diameter(geom)

        gravelius = perim / sqrt(4.0 * pi * area) if (0.0 < area) else None
        jaggedness = (perim * perim) / area if (0.0 < area) else None
        miller = (4.0 * pi * area) / (perim * perim) if (0.0 < perim) else None
        morton = (4.0 * area) / (pi * diamLen * diamLen) if (0.0 < diamLen) else None

        _, _, radius = ChrystalAlgorithm(geom).run()
        mbcArea = pi * radius * radius        
        areaCircDefect = area / mbcArea if (0.0 < mbcArea) else None

        return {
            'gravelius': gravelius,
            'jaggedness': jaggedness,
            'miller': miller,
            'morton': morton,
            'a_circ_def': areaCircDefect
            }



'''
Created on 25 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import exp, pi, sin
from t4gpd.commons.AngleLib import AngleLib


class Dogniaux(object):
    '''
    classdocs
    d'apres (Miguet, 2000; p. 170)
    '''
    SOLAR_CONSTANT = 1367.0
    '''
    http://atmospheres.gsfc.nasa.gov/climate/index.php?section=136
    The total solar irradiance (TSI), improperly called "solar constant" until a few years ago,
    has been found to change about 0.1% in an 11-year solar sunspot activity. The current most
    accurate TSI values from the Total Irradiance Monitor (TIM) on NASA's Solar Radiation and
    Climate Experiment (SORCE) is 1360.8 +/- 0.5 W/m2 during the 2008 solar minimum as compared
    to previous estimates of 1365.4 +/- 1.3 W/m2 established in the 1990s.
    '''
    TOTAL_SOLAR_IRRADIANCE = 1360.8

    @staticmethod
    def __opticalAirMass(solarAltitudeAngle):
        # https://en.wikipedia.org/wiki/Air_mass_%28astronomy%29
        # d'apres (Miguet, 2000; p. 164)
        assert (0 <= solarAltitudeAngle <= (pi / 2)), 'solarAltitudeAngle in radians!'
        '''
        return (1.0 / (sin(solarAltitudeAngle) + 0.15 * 
                       (AngleLib.toDegrees(solarAltitudeAngle) + 3.885) ** (-1.253)))
        return (1.0 / (sin(solarAltitudeAngle) + 0.50572 * 
                       (AngleLib.toDegrees(solarAltitudeAngle) + 6.07995) ** (-1.6364)))
        return 1.0 / sin(solarAltitudeAngle)
        '''
        return (1.0 / (sin(solarAltitudeAngle) + 0.50572 * 
                       (AngleLib.toDegrees(solarAltitudeAngle) + 6.07995) ** (-1.6364)))

    @staticmethod
    def __extinctionAtmosphericCoefficient(atmosphericTrouble, angstromTurbidity):
        if (angstromTurbidity < 0.075):
            return 0.1512 - 0.0262 * atmosphericTrouble
        elif (angstromTurbidity < 0.15):
            return 0.1656 - 0.0215 * atmosphericTrouble
        else:
            return 0.2021 - 0.0193 * atmosphericTrouble

    @staticmethod
    def directNormalIrradiance(solarAltitudeAngle):
        assert (0 <= solarAltitudeAngle <= (pi / 2)), 'solarAltitudeAngle in radians!'

        # TODO: Corresponding plotted diagram does not match (Miguet, 2000; p. 172)
        atmosphericTrouble = 3.0  # ~ urban area (Miguet, 2000; p. 166)
        angstromTurbidity = 0.15

        tmp = Dogniaux.__opticalAirMass(solarAltitudeAngle) * Dogniaux.__extinctionAtmosphericCoefficient(atmosphericTrouble, angstromTurbidity) * atmosphericTrouble
        return Dogniaux.SOLAR_CONSTANT * exp(-tmp) * sin(solarAltitudeAngle)



'''
Created on 20 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import sqrt
from shapely.geometry import LineString, MultiLineString, MultiPolygon, Point, Polygon
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class PointsDensifierLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def __densifyByCount(line, npts, blockid, contourid, distToTheSubstrate=None):
        distance = line.length / (npts - 1)
        return PointsDensifierLib.__densifyByDistance(line, distance, blockid,
                                                      contourid, distToTheSubstrate)

    @staticmethod
    def __densifyByDistance(line, distance, blockid, contourid, adjustableDist=True,
                            distToTheSubstrate=None):
        if not isinstance(line, LineString):
            raise IllegalArgumentTypeException(line, 'LineString')

        if adjustableDist:
            _d = line.length
            if (_d > distance):
                distance = _d / round(_d / distance)

        listOfCoords = line.coords

        result, nodeid, remainingDist = list(), 0, 0.0
        for i in range(1, len(listOfCoords)):
            a, b = listOfCoords[i - 1], listOfCoords[i]
            speedv = [ b[0] - a[0], b[1] - a[1] ]
            norm = sqrt(speedv[0] * speedv[0] + speedv[1] * speedv[1])
            if (0.0 < norm):
                speedv = [ speedv[0] / norm, speedv[1] / norm ]
                normalv = [speedv[1], -speedv[0], 0]

                k = remainingDist
                while (k <= norm):
                    samplePoint = Point([ a[0] + k * speedv[0], a[1] + k * speedv[1] ])

                    if not distToTheSubstrate is None:
                        samplePoint = Point([
                            samplePoint.x + distToTheSubstrate * normalv[0],
                            samplePoint.y + distToTheSubstrate * normalv[1]
                            ])

                    result.append({'node_id': ArrayCoding.encode([blockid, contourid, nodeid]),
                                   'motion_dir': ArrayCoding.encode(speedv),
                                   'normal_vec': ArrayCoding.encode(normalv),
                                   'geometry': samplePoint})
                    k += distance
                    nodeid += 1

                remainingDist = k - norm
        return result

    @staticmethod
    def __densifyByCurvilinearAbscissa(line, curvAbsc, blockid, contourid, distToTheSubstrate=None):
        if not isinstance(line, LineString):
            raise IllegalArgumentTypeException(line, 'LineString')

        listOfCoords = line.coords
        result, nodeid = list(), 0
        for i in range(1, len(listOfCoords)):
            a, b = listOfCoords[i - 1], listOfCoords[i]
            speedv = [ b[0] - a[0], b[1] - a[1] ]
            norm = sqrt(speedv[0] * speedv[0] + speedv[1] * speedv[1])
            if (0.0 < norm):
                speedv = [ speedv[0] / norm, speedv[1] / norm ]
                normalv = [speedv[1], -speedv[0], 0]

                ls = LineString([a, b])
                for _curvAbsc in curvAbsc:
                    samplePoint = ls.interpolate(_curvAbsc, normalized=True)

                    if not distToTheSubstrate is None:
                        samplePoint = Point([
                            samplePoint.x + distToTheSubstrate * normalv[0],
                            samplePoint.y + distToTheSubstrate * normalv[1]
                            ])

                    result.append({'node_id': ArrayCoding.encode([blockid, contourid, nodeid]),
                                   'motion_dir': ArrayCoding.encode(speedv),
                                   'normal_vec': ArrayCoding.encode(normalv),
                                   'geometry': samplePoint})
                    nodeid += 1

        return result

    @staticmethod
    def densifyByCount(geom, npts, blockid, contourid=0, distToTheSubstrate=None):
        if isinstance(geom, LineString):
            result = PointsDensifierLib.__densifyByCount(geom, npts, blockid, contourid,
                                                         distToTheSubstrate)

        elif isinstance(geom, Polygon):
            result = PointsDensifierLib.__densifyByCount(geom.exterior, npts, blockid, contourid,
                                                         distToTheSubstrate)
            for g in geom.interiors:
                contourid += 1
                result += PointsDensifierLib.__densifyByCount(g, npts, blockid, contourid,
                                                              distToTheSubstrate)

        elif isinstance(geom, MultiLineString) or isinstance(geom, MultiPolygon):
            result = list()
            for g in geom.geoms:
                result += PointsDensifierLib.__densifyByCount(g, npts, blockid, contourid,
                                                              distToTheSubstrate)
                contourid += 1

        else:
            result = []

        return result

    @staticmethod
    def densifyByDistance(geom, distance, blockid, contourid=0, adjustableDist=True,
                          distToTheSubstrate=None):
        if isinstance(geom, LineString):
            result = PointsDensifierLib.__densifyByDistance(
                geom, distance, blockid, contourid, adjustableDist, distToTheSubstrate)

        elif isinstance(geom, Polygon):
            result = PointsDensifierLib.__densifyByDistance(
                geom.exterior, distance, blockid, contourid, adjustableDist, distToTheSubstrate)
            for g in geom.interiors:
                contourid += 1
                result += PointsDensifierLib.__densifyByDistance(
                    g, distance, blockid, contourid, adjustableDist, distToTheSubstrate)

        elif isinstance(geom, MultiLineString) or isinstance(geom, MultiPolygon):
            result = list()
            for g in geom.geoms:
                result += PointsDensifierLib.densifyByDistance(
                    g, distance, blockid, contourid, adjustableDist, distToTheSubstrate)
                contourid += 1

        else:
            result = []

        return result

    @staticmethod
    def densifyByCurvilinearAbscissa(geom, curvAbsc, blockid, contourid=0, distToTheSubstrate=None):
        if isinstance(geom, LineString):
            result = PointsDensifierLib.__densifyByCurvilinearAbscissa(
                geom, curvAbsc, blockid, contourid, distToTheSubstrate)

        elif isinstance(geom, Polygon):
            result = PointsDensifierLib.__densifyByCurvilinearAbscissa(
                geom.exterior, curvAbsc, blockid, contourid, distToTheSubstrate)
            for g in geom.interiors:
                contourid += 1
                result += PointsDensifierLib.__densifyByCurvilinearAbscissa(
                    g, curvAbsc, blockid, contourid, distToTheSubstrate)

        elif isinstance(geom, MultiLineString) or isinstance(geom, MultiPolygon):
            result = list()
            for g in geom.geoms:
                result += PointsDensifierLib.densifyByCurvilinearAbscissa(
                    g, curvAbsc, blockid, contourid, distToTheSubstrate)
                contourid += 1

        else:
            result = []

        return result



'''
Created on 22 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely.geometry import Point
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class MoveSensorsAwayFromSurface(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, normalFieldname, dist=1e-6):
        '''
        Constructor
        '''
        assert isinstance(sensorsGdf, GeoDataFrame), 'sensorsGdf must be a GeoDataFrame'
        assert (normalFieldname in sensorsGdf), 'f{normalFieldname} must be a sensorsGdf field name'
        self.sensorsGdf = sensorsGdf
        self.normalFieldname = normalFieldname
        self.dist = dist

    def runWithArgs(self, row):
        p = row.geometry
        assert isinstance(p, Point), 'sensorsGdf must be a GeoDataFrame of Points'
        n = row[self.normalFieldname]

        return { 'geometry': Point([
            p.x + self.dist * n[0],
            p.y + self.dist * n[1],
            p.z + self.dist * n[2],
            ]) }



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 30 17:39:51 2021

@author: xlaffaille
"""
import math


def set_mist_optimized(tdb, tr, v, rh, met, clo, vapor_pressure, wme, body_surface_area,
                       patm, fa_eff, p_mist):  # fa_eff et p_mist ajoute par Xenia
    # Initial variables as defined in the ASHRAE 55-2017
    air_velocity = max(v, 0.1)
    k_clo = 0.25
    body_weight = 69.9
    met_factor = 58.2
    sbc = 0.000000056697  # Stefan-Boltzmann constant (W/m2K4)
    c_sw = 170  # driving coefficient for regulatory sweating
    c_dil = 120  # driving coefficient for vasodilation
    c_str = 0.5  # driving coefficient for vasoconstriction

    temp_skin_neutral = 33.7
    temp_core_neutral = 36.8
    temp_body_neutral = 36.49
    skin_blood_flow_neutral = 6.3

    temp_skin = temp_skin_neutral
    temp_core = temp_core_neutral
    skin_blood_flow = skin_blood_flow_neutral
    alfa = 0.1  # fractional skin mass
    e_sk = 0.1 * met  # total evaporative heat loss, W

    pressure_in_atmospheres = patm / 101325
    length_time_simulation = 60  # length time simulation
    r_clo = 0.155 * clo  # thermal resistance of clothing, C M^2 /W

    f_a_cl = 1.0 + 0.15 * clo  # increase in body surface area due to clothing
    lr = 2.2 / pressure_in_atmospheres  # Lewis ratio
    rm = met * met_factor  # metabolic rate
    m = met * met_factor

    if clo <= 0:
        w_crit = 0.38 * pow(air_velocity, -0.29)  # evaporative efficiency
        i_cl = 1.0  # thermal resistance of clothing, clo
    else:
        w_crit = 0.59 * pow(air_velocity, -0.08)
        i_cl = 0.45

    # h_cc corrected convective heat transfer coefficient
    h_cc = 3.0 * pow(pressure_in_atmospheres, 0.53)
    # h_fc forced convective heat transfer coefficient, W/(m2 C)
    h_fc = 8.600001 * pow((air_velocity * pressure_in_atmospheres), 0.53)
    h_cc = max(h_cc, h_fc)

    c_hr = 4.7  # linearized radiative heat transfer coefficient
    CTC = c_hr + h_cc
    r_a = 1.0 / (f_a_cl * CTC)  # resistance of air layer to dry heat
    t_op = (c_hr * tr + h_cc * tdb) / CTC  # operative temperature

    # initialize some variables
    dry = 0
    p_wet = 0
    _set_mist = 0  # ajoute _mist au _set par Xenia

    for TIM in range(length_time_simulation):

        iteration_limit = 150
        # t_cl temperature of the outer surface of clothing
        t_cl = (r_a * temp_skin + r_clo * t_op) / (r_a + r_clo)  # initial guess
        n_iterations = 0
        tc_converged = False

        while not tc_converged:

            c_hr = 4.0 * sbc * ((t_cl + tr) / 2.0 + 273.15) ** 3.0 * 0.72
            CTC = c_hr + h_cc
            r_a = 1.0 / (f_a_cl * CTC)
            t_op = (c_hr * tr + h_cc * tdb) / CTC
            t_cl_new = (r_a * temp_skin + r_clo * t_op) / (r_a + r_clo)
            if abs(t_cl_new - t_cl) <= 0.01:
                tc_converged = True
            t_cl = t_cl_new
            n_iterations += 1

            if n_iterations > iteration_limit:
                raise StopIteration("Max iterations exceeded")

        dry = (temp_skin - t_op) / (r_a + r_clo)  # total sensible heat loss, W
        # h_fcs rate of energy transport between core and skin, W
        h_fcs = (temp_core - temp_skin) * (5.28 + 1.163 * skin_blood_flow)
        q_res = 0.0023 * m * (44.0 - vapor_pressure)  # heat loss due to respiration
        CRES = 0.0014 * m * (34.0 - tdb)
        s_core = m - h_fcs - q_res - CRES - wme  # rate of energy storage in the core
        s_skin = h_fcs - dry - e_sk  # rate of energy storage in the skin
        TCSK = 0.97 * alfa * body_weight
        TCCR = 0.97 * (1 - alfa) * body_weight
        DTSK = (s_skin * body_surface_area) / (TCSK * 60.0)  # C per minute
        DTCR = s_core * body_surface_area / (TCCR * 60.0)
        temp_skin = temp_skin + DTSK
        temp_core = temp_core + DTCR
        t_body = alfa * temp_skin + (1 - alfa) * temp_core  # mean body temperature, C
        # sk_sig thermoregulatory control signal from the skin
        sk_sig = temp_skin - temp_skin_neutral
        warms = (sk_sig > 0) * sk_sig  # vasodialtion signal
        colds = ((-1.0 * sk_sig) > 0) * (-1.0 * sk_sig)  # vasoconstriction signal
        # c_reg_sig thermoregulatory control signal from the skin, C
        c_reg_sig = temp_core - temp_core_neutral
        # c_warm vasodilation signal
        c_warm = (c_reg_sig > 0) * c_reg_sig
        # c_cold vasoconstriction signal
        c_cold = ((-1.0 * c_reg_sig) > 0) * (-1.0 * c_reg_sig)
        BDSIG = t_body - temp_body_neutral
        WARMB = (BDSIG > 0) * BDSIG
        skin_blood_flow = (skin_blood_flow_neutral + c_dil * c_warm) / (
            1 + c_str * colds
        )
        if skin_blood_flow > 90.0:
            skin_blood_flow = 90.0
        if skin_blood_flow < 0.5:
            skin_blood_flow = 0.5
        REGSW = c_sw * WARMB * math.exp(warms / 10.7)
        if REGSW > 500.0:
            REGSW = 500.0
        e_rsw = 0.68 * REGSW  # heat lost by vaporization sweat
        r_ea = 1.0 / (lr * f_a_cl * h_cc)  # evaporative resistance air layer
        r_ecl = r_clo / (lr * i_cl)
        # e_max = maximum evaporative capacity
        e_max = (
            math.exp(18.6686 - 4030.183 / (temp_skin + 235.0)) - vapor_pressure
        ) / (r_ea + r_ecl)
        p_rsw = e_rsw / e_max  # ratio heat loss sweating to max heat loss sweating
        p_wet = 0.06 + 0.94 * p_rsw  # skin wetness
        e_diff = p_wet * e_max - e_rsw  # vapor diffusion through skin
        if p_wet > w_crit:
            p_wet = w_crit
            p_rsw = w_crit / 0.94
            e_rsw = p_rsw * e_max
            e_diff = 0.06 * (1.0 - p_rsw) * e_max
        if e_max < 0:
            e_diff = 0
            e_rsw = 0
            p_wet = w_crit
        e_sk = (
            e_rsw + e_diff
        )  # total evaporative heat loss sweating and vapor diffusion
        MSHIV = 19.4 * colds * c_cold
        m = rm + MSHIV
        alfa = 0.0417737 + 0.7451833 / (skin_blood_flow + 0.585417)
        
    #################################################################### Debut Xenia
    # saturated water vapor pressure at the clothing temperature
    p_clo = math.exp(18.956 - 4030.183 / (t_cl + 235)) 
    p_clo = p_clo / 10  # [kPa]
  
    # rate of total evaporative heat loss from the skin at misting
    e_mist = fa_eff * (1 - p_wet) * p_mist * (p_clo - vapor_pressure) / r_ea
  
    # sum of skin wetness at clothed nodes Esk+Emist - total evaporative heat loss from the skin in misting environment
    e_clo = e_sk + e_mist
  
    # total heat loss from skin, W
    hsk = dry + e_clo  
    #################################################################### Fin Xenia
    
    W = p_wet
    PSSK = math.exp(18.6686 - 4030.183 / (temp_skin + 235.0))
    CHRS = c_hr
    if met < 0.85:
        CHCS = 3.0
    else:
        CHCS = 5.66 * (met - 0.85) ** 0.39
    if CHCS < 3.0:
        CHCS = 3.0
    CTCS = CHCS + CHRS
    RCLOS = 1.52 / ((met - wme / met_factor) + 0.6944) - 0.1835
    RCLS = 0.155 * RCLOS
    FACLS = 1.0 + k_clo * RCLOS
    FCLS = 1.0 / (1.0 + 0.155 * FACLS * CTCS * RCLOS)
    IMS = 0.45
    ICLS = IMS * CHCS / CTCS * (1 - FCLS) / (CHCS / CTCS - FCLS * IMS)
    RAS = 1.0 / (FACLS * CTCS)
    REAS = 1.0 / (lr * FACLS * CHCS)
    RECLS = RCLS / (lr * ICLS)
    HD_S = 1.0 / (RAS + RCLS)
    HE_S = 1.0 / (REAS + RECLS)

    delta = 0.0001
    dx = 100.0
    set_mist_old = round(temp_skin - hsk / HD_S, 2)
    while abs(dx) > 0.01:  # ajoute _mist au _set par Xenia
        err_1 = (
            hsk
            -HD_S * (temp_skin - set_mist_old)
            -W
            * HE_S
            * (PSSK - 0.5 * (math.exp(18.6686 - 4030.183 / (set_mist_old + 235.0))))
        )
        err_2 = (
            hsk
            -HD_S * (temp_skin - (set_mist_old + delta))
            -W
            * HE_S
            * (PSSK - 0.5 * (math.exp(18.6686 - 4030.183 / (set_mist_old + delta + 235.0))))
        )
        _set_mist = set_mist_old - delta * err_1 / (err_2 - err_1)
        dx = _set_mist - set_mist_old
        set_mist_old = _set_mist

    return _set_mist



'''
Created on 8 nov. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import read_file
from os import access, R_OK
from os.path import isfile
from pathlib import Path
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from zipfile import is_zipfile, ZipFile


class ZipLoader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, zipInputFile, layernames=None, bbox=None, mask=None):
        '''
        Constructor
        '''
        self.zipInputFile = zipInputFile
        if not (isfile(self.zipInputFile) and access(self.zipInputFile, R_OK)):
            raise IllegalArgumentTypeException(zipInputFile, 'readable filename')
        if not is_zipfile(self.zipInputFile):
            raise IllegalArgumentTypeException(zipInputFile, 'Zip filename')

        self.layernames = layernames
        self.bbox = bbox
        self.mask = mask

    def __getTypeOfDataFrame(self, extension):
        if extension in ['.geojson', '.gpkg', '.shp']:
            return 'GeoDataFrame'
        return None

    def run(self):
        if self.layernames is None:
            layers = []
        else:
            layers = { _layername: None for _layername in self.layernames}

        with ZipFile(self.zipInputFile, 'r') as _archive:
            for _filename in _archive.namelist():
                _extension = Path(_filename).suffix.lower()
                _basename = Path(_filename).stem
                _tod = self.__getTypeOfDataFrame(_extension)

                if _tod is not None:
                    if self.layernames is None:
                        layers.append(_basename)

                    elif _basename in self.layernames:
                        _gdf = read_file(f'{self.zipInputFile}!{_filename}',
                                         bbox=self.bbox, mask=self.mask)
                        layers[_basename] = _gdf

        if self.layernames is None:
            if (0 < len(layers)):
                print(f'List of GeoDataFrame(s): {layers}')
            return None

        return [layers[_layername] for _layername in self.layernames]



'''
Created on 9 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import asarray, copy, dot
from numpy.random import shuffle
from shapely.geometry import MultiLineString, MultiPoint, Point
from sklearn.preprocessing import normalize
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid
from t4gpd.pyvista.commons.RayCastingIn3DLib import RayCastingIn3DLib


class RayCasting3D(AbstractGeoprocess):
    '''
    classdocs
    '''
    def __init__(self, gdfs, shootingDirs, viewpoints=None, normalFieldname=None,
                 pkFieldname=None, mc=None, maxRayLen=None, showHitPoints=False,
                 encode=False):
        '''
        Constructor
        '''
        assert isinstance(gdfs, (list, tuple)), 'gdfs must be a list of GeoDataFrames'
        assert all(isinstance(gdf, GeoDataFrame) for gdf in gdfs), 'gdfs must be a list of GeoDataFrames'

        self.shootingDirs = asarray(shootingDirs)
        assert 2 == self.shootingDirs.ndim, 'shootingDirs must be a list of 2D/3D vectors'
        self.shootingDirs = normalize(self.shootingDirs, norm='l2', copy=True) 

        assert (viewpoints is None) or isinstance(viewpoints, GeoDataFrame), 'viewpoints must be either None or a GeoDataFrame'
        assert (normalFieldname is None) or (normalFieldname in viewpoints), 'normalFieldname must be either None or a viewpoints field name'
        self.normalFieldname = normalFieldname

        if pkFieldname is not None:
            assert all((pkFieldname in gdf) for gdf in gdfs), f'{pkFieldname} must be a GeoDataFrame field name!'

        self.pkFieldname = pkFieldname
        '''
        https://kitware.github.io/vtk-examples/site/Python/
        https://pyscience.wordpress.com/2014/09/21/ray-casting-with-python-and-vtk-intersecting-linesrays-with-surface-meshes/

        Excerpt from https://www.kitware.com/ray-casting-ray-tracing-with-vtk/

        the vtkOBBTree class, which generates an oriented bounding-box (OBB) ‘tree’
        for a given geometry under a vtkPolyData object. Upon generation of this OBB 
        tree, the vtkOBBTree class allows us to perform intersection tests between
        the mesh and the lines of finite length, as well as intersection tests between
        different meshes. It can then return the point coordinates where intersections
        were detected, as well as the polydata cell IDs where the intersections occurred.
        '''
        masks = ToUnstructuredGrid(gdfs, fieldname=pkFieldname).run()
        self.bounds = masks.bounds

        self.obbTree = RayCastingIn3DLib.prepareVtkOBBTree(masks)

        maxRayLen = masks.length if maxRayLen is None else maxRayLen
        self.shootingDirs *= maxRayLen 

        self.showHitPoints = showHitPoints
        assert (mc is None) or (0 <= mc <= 1.0), 'mc parameter must be either None or in the range 0 to 1'
        self.mc = mc
        self.encode = encode

    '''
    @staticmethod
    def __clipWithBounds(rays, bounds):
        _rays = GeoDataFrame(data=[{'geometry': MultiLineString(rays)}])
        _rays = _rays.explode(ignore_index=True)
        _rays = ToUnstructuredGrid([_rays]).run()
        _rays = _rays.clip_box(*bounds)
        _rays = [_rays.cell_points(i) for i in range(_rays.n_cells)]
        return _rays
    '''

    def runWithArgs(self, row):
        geom = row.geometry
        if not isinstance(geom, Point):
            raise IllegalArgumentTypeException(geom, 'GeoDataFrame of Point')

        shootingDirs = copy(self.shootingDirs)
        if self.mc is not None:
            shuffle(shootingDirs)
            shootingDirs = shootingDirs[0:int(self.mc * len(shootingDirs))]

        if self.normalFieldname is not None:
            normalVector = row[self.normalFieldname]
            normalVector = ArrayCoding.decode(normalVector) if isinstance(normalVector, str) else normalVector
            shootingDirs = shootingDirs[ dot(shootingDirs, normalVector) > 0 ]

        srcPt = asarray(geom.coords[0])
        dstPts = srcPt + shootingDirs
        srcPts = srcPt.reshape(1, -1).repeat(len(dstPts), axis=0)

        rays, hitPoints, hitDists, hitGids = RayCastingIn3DLib.mraycastObbTree(
            self.obbTree, srcPts, dstPts, self.pkFieldname)

        # rays = RayCasting3D.__clipWithBounds(rays, self.bounds)

        if self.pkFieldname is None:
            return {
                'geometry': MultiPoint(hitPoints) if (self.showHitPoints) else MultiLineString(rays),
                'n_rays': len(shootingDirs),
                'hitDists': ArrayCoding.encode(hitDists) if self.encode else hitDists,
                'inf_ratio': len([d for d in hitDists if (d == RayCastingIn3DLib.INFINITY)]) / len(hitDists)
                }

        return {
            'geometry': MultiPoint(hitPoints) if (self.showHitPoints) else MultiLineString(rays),
            'n_rays': len(shootingDirs),
            'hitDists': ArrayCoding.encode(hitDists) if self.encode else hitDists,
            'hitGids': ArrayCoding.encode(hitGids) if self.encode else hitGids,
            'inf_ratio': len([d for d in hitDists if (d == RayCastingIn3DLib.INFINITY)]) / len(hitDists)
            }



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.MyNode import MyNode


class MyEdge(object):
    '''
    classdocs
    '''

    def __init__(self, startNode, endNode):
        '''
        Constructor
        '''
        if isinstance(startNode, list):
            startNode = MyNode(startNode)
        if isinstance(endNode, list):
            endNode = MyNode(endNode)
        startHash = startNode.__hash__()
        endHash = endNode.__hash__()

        if startHash < endHash:
            self.startNode, self.endNode = [ startNode, endNode ]
            self.startHash, self.endHash = [ startHash, endHash ]
        else:
            self.startNode, self.endNode = [ endNode, startNode ]
            self.startHash, self.endHash = [ endHash, startHash ]

        self.hash = None

    def __eq__(self, other):
        # print 'test MyEdge equality: %s =?= %s' % (self, other)
        return isinstance(other, MyEdge) and (self.startNode == other.startNode) and (self.endNode == other.endNode) 

    def __hash__(self):
        if self.hash is None:
            self.hash = 17
            self.hash = 37 * self.hash + self.startHash
            self.hash = 37 * self.hash + self.endHash
        # print 'hash%s = %d' % (self, self.hash)
        return self.hash

    @staticmethod
    def isnumeric(x):
        return isinstance(x, (int, float))

    def __repr__(self):
        return '[%s --- %s]' % (self.startNode, self.endNode)



'''
Created on 24 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from pandas import DataFrame
from shapely.geometry import Point
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STSnappingPointsOnLines2(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, pointsGdf, linesGdf, wayPointsGdf, stepCountFieldname, tagName,
                 wayPointsIdFieldname):
        '''
        Constructor
        '''
        if not isinstance(pointsGdf, DataFrame):
            raise IllegalArgumentTypeException(pointsGdf, 'DataFrame')
        self.pointsGdf = pointsGdf

        if not isinstance(linesGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(linesGdf, 'GeoDataFrame')
        self.linesGdf = linesGdf
        self.spatialIdx = linesGdf.sindex
        if (1 != len(self.linesGdf)):
            raise Exception('%s is not a one-row GeoDataFrame!' % (self.linesGdf))

        if not isinstance(wayPointsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(wayPointsGdf, 'GeoDataFrame')
        self.wayPointsGdf = wayPointsGdf

        if stepCountFieldname not in pointsGdf:
            raise Exception('%s is not a relevant field name!' % (stepCountFieldname))
        self.stepCountFieldname = stepCountFieldname

        if tagName not in pointsGdf:
            raise Exception('%s is not a relevant field name!' % (tagName))
        self.tagName = tagName

        if wayPointsIdFieldname not in wayPointsGdf:
            raise Exception('%s is not a relevant field name!' % (wayPointsIdFieldname))
        self.wayPointsIdFieldname = wayPointsIdFieldname

    def run(self):
        pathGeom = self.linesGdf.geometry.squeeze()

        # 1st STEP: PROJECT WAYPOINTS ONTO THE LINES
        projWayPoints = dict()
        for _, wayPoint in self.wayPointsGdf.iterrows():
            wayPointId = wayPoint[self.wayPointsIdFieldname]
            wayPointCurvDist = pathGeom.project(wayPoint.geometry)
            wayPointGeom = pathGeom.interpolate(wayPointCurvDist)

            projWayPoints[wayPointId] = {
                'gid': wayPointId,
                'curv_absc': wayPointCurvDist,
                'geometry': wayPointGeom
                }

        # 2nd STEP: READ MEASURE POINTS AND IDENTIFY THE POTENTIAL WAYPOINTS
        self.pointsGdf = self.pointsGdf.sort_values(self.stepCountFieldname)

        prevStartStepCount, prevCurvAbsc = 0, 0.0
        _pointsGdf = self.pointsGdf[
            (~self.pointsGdf[self.tagName].isin(['NA', '"NA"'])) & 
            (~self.pointsGdf[self.tagName].isnull())
            ]

        for _, point in _pointsGdf.iterrows():
            tagName = point[self.tagName]
            tagName = int(tagName.replace('"', '')) if isinstance(tagName, str) else tagName
            stepCount = point[self.stepCountFieldname]

            if tagName in projWayPoints:
                projWayPoints[tagName]['startStepCount'] = prevStartStepCount
                projWayPoints[tagName]['stopStepCount'] = stepCount
                projWayPoints[tagName]['nPoints'] = 0
                projWayPoints[tagName]['deltaL'] = projWayPoints[tagName]['curv_absc'] - prevCurvAbsc
                prevStartStepCount = stepCount + 1
                prevCurvAbsc = projWayPoints[tagName]['curv_absc']
            else:
                raise Exception('Unreachable instruction!')

        wayPointIdMax = 1 + max([ v['gid'] for v in projWayPoints.values() ])
        stepCountMax = max(self.pointsGdf[self.stepCountFieldname])

        projWayPoints[wayPointIdMax] = {
            'gid': wayPointIdMax,
            'curv_absc': pathGeom.length,
            'geometry': Point((list(pathGeom.coords)[-1])),
            'startStepCount': prevStartStepCount,
            'stopStepCount': stepCountMax,
            'nPoints': 0,
            'deltaL': pathGeom.length - prevCurvAbsc
            }

        # 3rd STEP: COUNT MEASURE POINTS PER PAIR OF WAYPOINTS
        for _, point in self.pointsGdf.iterrows():
            stepCount = point[self.stepCountFieldname]
            for tagName in projWayPoints.keys():
                if (projWayPoints[tagName]['startStepCount'] <= stepCount <= projWayPoints[tagName]['stopStepCount']):
                    projWayPoints[tagName]['nPoints'] += 1

        for tagName in list(projWayPoints.keys()):
            if (0 == projWayPoints[tagName]['nPoints']):
                del(projWayPoints[tagName])
            else:
                projWayPoints[tagName]['deltaL'] = projWayPoints[tagName]['deltaL'] / projWayPoints[tagName]['nPoints']

        # 4th STEP: POPULATE THE OUTPUT ROWS RESULT
        inputMeasurePoints = dict()
        for _, point in self.pointsGdf.iterrows():
            stepCount = point[self.stepCountFieldname]
            inputMeasurePoints[stepCount] = point

        initialStartStepCount = min(inputMeasurePoints.keys())
        currCurvAbsc, startStepCount, rows = 0, initialStartStepCount, []
        for tagName in sorted(projWayPoints.keys()):
            incCurvAbsc = projWayPoints[tagName]['deltaL']
            stopStepCount = projWayPoints[tagName]['stopStepCount']

            for stepCount in range(startStepCount, stopStepCount + 1):
                if initialStartStepCount < stepCount:
                    currCurvAbsc += incCurvAbsc
                row = inputMeasurePoints[stepCount].to_dict()
                projPoint = pathGeom.interpolate(currCurvAbsc)
                dist_to_l = row['geometry'].distance(projPoint) if ('geometry' in row) else None    
                row['geometry'] = projPoint
                row['dist_to_l'] = dist_to_l
                row['curv_absc'] = currCurvAbsc
                rows.append(row)

            startStepCount = stopStepCount + 1

        return GeoDataFrame(rows, crs=self.wayPointsGdf.crs)



'''
Created on 18 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import LineString
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class FromContourToNormalVector(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, magn=1.0):
        self.magn = magn

    def runWithArgs(self, row):
        geom = row.geometry
        assert GeomLib.isPolygonal(geom), 'The GeoDataFrame must be composed of (Multi)Polygon'
        c = GeomLib3D.centroid(geom).coords[0]
        n = GeomLib3D.getFaceNormalVector(geom)
        return { 'geometry': LineString([c, [c[i] + self.magn * n[i] for i in range(3)]]) }



'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arctan2, cos, isnan, mean, pi, sin, sqrt

from t4gpd.commons.ListUtilities import ListUtilities


class PolarCartesianCoordinates(object):
    '''
    classdocs
    '''

    @staticmethod
    def fromCartesianToPolarCoordinates(x, y):
        r = sqrt((x * x) + (y * y))
        if (0 == x):
            if (0 < y):
                azimuth = pi / 2
            elif (0 > y):
                azimuth = (3 * pi) / 2
            else:
                azimuth = float('nan')
        else:
            azimuth = arctan2(y, x)
            if (0 > azimuth):
                azimuth += 2 * pi
        return [float(r), float(azimuth)]

    @staticmethod
    def fromPolarToCartesianCoordinates(r, azimuth):
        return [r * cos(azimuth), r * sin(azimuth)]

    @staticmethod
    def fromRayLengthsToPolarCoordinates(rayLengths):
        if isinstance(rayLengths, list) and (0 < len(rayLengths)):
            angularOffset = (2 * pi) / len(rayLengths)
            return [[r, i * angularOffset] for i, r in enumerate(rayLengths)]
        return None

    @staticmethod
    def fromRayLengthsToCartesianCoordinates(rayLengths):
        if isinstance(rayLengths, list) and (0 < len(rayLengths)):
            angularOffset = (2 * pi) / len(rayLengths)
            return [PolarCartesianCoordinates.fromPolarToCartesianCoordinates(r, i * angularOffset) for i, r in enumerate(rayLengths)]
        return None

    @staticmethod
    def normalizeRayLengths(rayLengths, threshold):
        if isinstance(rayLengths, list) and (0 < len(rayLengths)):
            cartesianCoords = PolarCartesianCoordinates.fromRayLengthsToCartesianCoordinates(rayLengths)
            xmean = mean([xy[0] for xy in cartesianCoords])
            ymean = mean([xy[1] for xy in cartesianCoords])
            rmean, azimuthMean = PolarCartesianCoordinates.fromCartesianToPolarCoordinates(xmean, ymean)

            if isnan(azimuthMean) or (rmean < threshold):
                return rayLengths
            offset = int(round((len(rayLengths) * azimuthMean) / (2 * pi)))
            return ListUtilities.rotate(rayLengths, offset)
        return None



'''
Created on 17 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class IdentifyTheClosestPolyline(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, roadsGdf, roadsIdFieldname, defaultBuffDist=40.0):
        '''
        Constructor
        '''
        if not isinstance(roadsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(roadsGdf, 'GeoDataFrame')
        self.roadsGdf = roadsGdf
        self.spatialIdx = roadsGdf.sindex

        if roadsIdFieldname not in roadsGdf:
            raise Exception('%s is not a relevant field name!' % (roadsIdFieldname))
        self.roadsIdFieldname = roadsIdFieldname

        self.buffDist = defaultBuffDist

    def __sign(self, x):
        if (0 == x):
            return 0
        return -1 if (x < 0) else 1

    def runWithArgs(self, row):
        geom = row.geometry

        self.buffDist = 40.0
        minDist, nearestPoint, nearestRow = GeomLib.getNearestFeature(
            self.roadsGdf, self.spatialIdx, geom, self.buffDist)

        nearestRoad = nearestRow.geometry
        curvilinearAbscissa = nearestRoad.project(nearestPoint)

        otherPoint = nearestRoad.interpolate(curvilinearAbscissa + 1e-2)
        roadSide = self.__sign(GeomLib.zCrossProduct(
            otherPoint.coords[0], nearestPoint.coords[0], geom.coords[0]))

        return { 
            # 'geometry': LineString((geom, nearestPoint)),
            'road_id': nearestRow[self.roadsIdFieldname],
            'road_dist': minDist,
            'road_absc': curvilinearAbscissa,
            'road_side': roadSide
            }



'''
Created on 5 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from numpy import arange, array, ndarray
from pyvista import MultiBlock, MultipleLines, PolyData 
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class ToPolyData(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdfs, fieldname=None):
        '''
        Constructor
        '''
        if not isinstance(gdfs, (list, tuple, ndarray)):
            raise IllegalArgumentTypeException(gdfs, 'list or tuple of GeoDataFrames')
        self.gdfs = gdfs
        self.fieldname = fieldname

    def __fromGeomToPolyData(self, geom):
        # https://vtk.org/wp-content/uploads/2015/04/file-formats.pdf
        if ('Point' == geom.geom_type):
            point = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.coords]
            point = array(point)
            return [ PolyData(point) ]

        elif ('LineString' == geom.geom_type):
            points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.coords]
            points = array(points)
            return [ MultipleLines(points) ]

        elif ('Polygon' == geom.geom_type):
            if (GeomLib.isConvex(geom) or GeomLib.isHoled(geom)):
                points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.exterior.coords]
                npoints = len(points)
                faces = [npoints] + arange(npoints).tolist()
                points, faces = array(points), array(faces)
                return [ PolyData(points, faces) ]
            else:
                result = []
                triangles = GeomLib.fromPolygonToListOfTriangles(geom)
                for triangle in triangles:
                    points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in triangle.exterior.coords]
                    npoints = len(points)
                    faces = [npoints] + arange(npoints).tolist()
                    points, faces = array(points), array(faces)
                    result.append(PolyData(points, faces))
                return result

        raise Exception('Do not handle GeometryCollection!')

    def __fromGeoDataFrameToDictOfPolyDatas(self, gdf, rank):
        gdf = gdf.explode(ignore_index=True)
        # print(gdf.columns)
        result = {}
        for idx, row in gdf.iterrows():
            geom = row.geometry
            polyDatas = self.__fromGeomToPolyData(geom)

            # we can add some attributes to the current PolyData
            if self.fieldname is not None:
                attrValue = row[self.fieldname] if self.fieldname in gdf else float('nan')
                for polyData in polyDatas:
                    polyData.cell_data[self.fieldname] = attrValue

            for i, polyData in enumerate(polyDatas):
                result[f'{rank}_{idx}_{i}'] = polyData
        return result

    def run(self):
        result = {}
        rank = 0
        for rank, gdf in enumerate(self.gdfs):
            result.update(self.__fromGeoDataFrameToDictOfPolyDatas(gdf, rank))
        polyBlocks = MultiBlock(result)
        polyGrid = polyBlocks.combine()
        return polyGrid



'''
Created on 25 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import exp, pi
from t4gpd.commons.AngleLib import AngleLib


class Perraudeau(object):
    '''
    classdocs
    d'apres (Miguet, 2000; p. 170)
    '''

    @staticmethod
    def directNormalIrradiance(solarAltitudeAngle):
        assert (0 <= solarAltitudeAngle <= (pi / 2)), 'solarAltitudeAngle in radians!'

        return 1000.0 * (1.0 - exp(-0.055 * AngleLib.toDegrees(solarAltitudeAngle)))



'''
Created on 6 juil. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import re


class CSVLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def readLexeme(s, decimalSep='.'):
        if ('.' != decimalSep):
            s = s.replace(decimalSep, '.')
        try:
            nb = float(s)
            if nb.is_integer():
                return int(nb)
            return nb
        except ValueError:
            pass
        try:
            import unicodedata
            nb = unicodedata.numeric(s)
            if nb.is_integer():
                return int(nb)
            return nb
        except (TypeError, ValueError):
            pass
        return s

    @staticmethod
    def read(inputFile, fieldSep=',', decimalSep='.'):
        rows, outputFields = [], []

        with open(inputFile, 'r') as f:
            for nline, line in enumerate(f, start=1):
                # values = line.strip().split(self.fieldSep)
                values = re.split(fieldSep, line.strip())
                values = [value.strip() for value in values]

                if (1 == nline):
                    for fieldName in values:
                        fieldName = fieldName.replace('"', '')
                        if (0 == len(fieldName)):
                            fieldName = 'gid'
                        outputFields.append(fieldName)
                else:
                    row = dict()
                    for i, value in enumerate(values):
                        value = value.replace('"', '')
                        row[outputFields[i]] = CSVLib.readLexeme(value, decimalSep)
                    rows.append(row)
        return rows



# 	PET calculation after the LadyBug plugin (retrieved on Djordje Spasic's github :
# 	https://github.com/stgeorges/ladybug/commit/b0c2ea970252b62d22bf0e35d739db7f385a3f26)
#
# 	2017.11.10 by Edouard Walther and Quentin Goestschel:
# 		- fixed the error on the reference environment (see paper)
# 	
import math as math
import matplotlib.pyplot as plt
import numpy as np

# Input data #
# Environment constants  #
po = 1013.25  # atmospheric pressure [hPa]
p = 1013.25  # real pressure [hPa]
rob = 1.06  # Blood density kg/L
cb = 3640.0  # Blood specific heat [j/kg/k]
emsk = 0.99  # Skin emissivity [-]
emcl = 0.95  # Clothes emissivity [-]
Lvap = 2.42 * math.pow(10.0, 6.0)  # Latent heat of evaporation [J/Kg]
sigm = 5.67 * math.pow(10.0, -8.0)  # Stefan-Boltzmann constant [W/(m2*K^(-4))]
cair = 1010.0  # Air specific heat  [J./kg/K-]
rdsk = 0.79 * math.pow(10.0, 7.0)  # Skin diffusivity
rdcl = 0.0  # Clothes diffusivity

sex = 1
pos = 1
age = 35
mbody = 75  # Subject weight[kg]
ht = 1.80  # Subject size[m]
Adu = 0.203 * math.pow(mbody, 0.425) * math.pow(ht, 0.725)  # Dubois body area
bodyPosition = "standing"
feff = 0.725
sex = "male"

# Initialisation of the temperature set values
tc_set = 36.6
tsk_set = 34
tbody_set = 0.1 * tsk_set + 0.9 * tc_set


# calcul de la reaction metabolique
def system(ta, tmrt, rh, v_air, M, Icl):
	vpa = rh  # ACHTUNG
	# Area parameters of the body: #
	if Icl < 0.03:
		Icl = 0.02
	icl = Icl  # [clo] Clothing level
	eta = 0.0  # Body efficiency
	# Calculation of the Burton coefficient, k = 0.31 for Hoeppe:
	fcl = 1 + (0.31 * icl)  # Increasment of the exchange area depending on the clothing level:
	if bodyPosition == "sitting":
		feff = 0.696
	elif bodyPosition == "standing":
		feff = 0.725
	elif bodyPosition == "crouching":
		feff = 0.67
	facl = (173.51 * icl - 2.36 - 100.76 * icl * icl + 19.28 * math.pow(icl, 3.0)) / 100.0

	# Basic metabolism for men and women in [W/m2] #
	# Attribution of internal energy depending on the sex of the subject
	if sex == "male":
		met_base = 3.45 * math.pow(mbody, 0.75) * (1.0 + 0.004 * (30.0 - age) + 0.01 * (ht * 100.0 / math.pow(mbody, 1.0 / 3.0) - 43.4))
	else:
		met_base = 3.19 * math.pow(mbody, 0.75) * (1.0 + 0.004 * (30.0 - age) + 0.018 * (ht * 100.0 / math.pow(mbody, 1.0 / 3.0) - 42.1))
	# Source term : metabolic activity
	he = M + met_base
	h = he * (1.0 - eta)

	# Respiratory energy losses #
	# Expired air temperature calculation:
	texp = 0.47 * ta + 21.0

	# Pulmonary flow rate
	rtv = he * 1.44 * math.pow(10.0, -6.0)

	# Sensible heat energy loss:
	Cres = cair * (ta - texp) * rtv

	# Latent heat energy loss:
	vpexp = 6.11 * math.pow(10.0, 7.45 * texp / (235.0 + texp))  # Partial pressure of the breathing air
	Eres = 0.623 * Lvap / p * (vpa - vpexp) * rtv
	# total breathing heat loss
	qresp = (Cres + Eres)

	c = [0 for i in range(11)]
	tcore = [0 for i in range(7)]  # Core temperature list
	hc = 2.67 + 6.5 * math.pow(v_air, 0.67)  # Convection coefficient
	hc = hc * math.pow(p / po, 0.55)  # Correction with pressure

	# Clothed fraction of the body approximation #
	rcl = icl / 6.45  # conversion in m2.K/W
	y = 0
	if facl > 1.0:
		facl = 1.0
		rcl = icl / 6.45  # conversion clo --> m2.K/W
	# y : equivalent clothed height of the cylinder
	# High clothing level : all the height of the cylinder is covered
	if icl >= 2.0:
		y = 1.0
	if icl > 0.6 and icl < 2.0:
		y = (ht - 0.2) / ht
	if icl <= 0.6 and icl > 0.3:
		y = 0.5
	if icl <= 0.3 and icl > 0.0:
		y = 0.1
	# calculation of the closing radius depending on the clothing level (6.28 = 2* pi !)
	r2 = Adu * (fcl - 1.0 + facl) / (6.28 * ht * y)  # External radius
	r1 = facl * Adu / (6.28 * ht * y)  # Internal radius
	di = r2 - r1
	# clothed surface
	Acl = Adu * facl + Adu * (fcl - 1.0)
	# skin temperatures
	for j in range(1, 7):
		tsk = tsk_set
		count1 = 0
		tcl = (ta + tmrt + tsk) / 3.0  # Average value between the temperatures to estimate Tclothes
		enbal2 = 0.0
		while True:
			for count2 in range(1, 100):
				# Estimation of the radiation losses
				rclo2 = emcl * sigm * (math.pow(tcl + 273.2, 4.0) - math.pow(tmrt + 273.2, 4.0)) * feff
				# Calculation of the thermal resistance of the body:
				htcl = (6.28 * ht * y * di) / (rcl * math.log(r2 / r1) * Acl)
				tsk = (hc * (tcl - ta) + rclo2) / htcl + tcl  # Skin temperature calculation

				# Radiation losses #
				Aeffr = Adu * feff  # Effective radiative area depending on the position of the subject
				# For bare skin area:
				rbare = Aeffr * (1.0 - facl) * emsk * sigm * (math.pow(tmrt + 273.2, 4.0) - math.pow(tsk + 273.2, 4.0))
				# For dressed area:
				rclo = feff * Acl * emcl * sigm * (math.pow(tmrt + 273.2, 4.0) - math.pow(tcl + 273.2, 4.0))
				rsum = rbare + rclo  # [W]

				# Convection losses #
				cbare = hc * (ta - tsk) * Adu * (1.0 - facl)
				cclo = hc * (ta - tcl) * Acl
				csum = cbare + cclo  # [W]

				# Calculation of the Terms of the second order polynomial :
				K_blood = Adu * rob * cb
				c[0] = h + qresp
				c[2] = tsk_set / 2 - 0.5 * tsk
				c[3] = 5.28 * Adu * c[2]
				c[4] = 13.0 / 625.0 * K_blood
				c[5] = 0.76275 * K_blood
				c[6] = c[3] - c[5] - tsk * c[4]
				c[7] = -c[0] * c[2] - tsk * c[3] + tsk * c[5]
				c[9] = 5.28 * Adu - 0.76275 * K_blood - 13.0 / 625.0 * K_blood * tsk
				# discriminant #1 (b^2 - 4*a*c)
				c[10] = math.pow((5.28 * Adu - 0.76275 * K_blood - 13.0 / 625.0 * K_blood * tsk), 2) - 4.0 * c[4] * (c[5] * tsk - c[0] - 5.28 * Adu * tsk)
				# discriminant #2 (b^2 - 4*a*c)
				c[8] = c[6] * c[6] - 4.0 * c[4] * c[7]
				if tsk == tsk_set:
					tsk = tsk_set + 0.01
				# Calculation of Tcore[]:
				# case 6 : Set blood flow only
				tcore[6] = (h + qresp) / (5.28 * Adu + K_blood * 6.3 / 3600.0) + tsk
				# cas 2 : Set blood flow + regulation
				tcore[2] = (h + qresp) / (5.28 * Adu + K_blood * 6.3 / 3600.0 / (1.0 + 0.5 * (tsk_set - tsk))) + tsk
				# case 3 : Maximum blood flow only
				tcore[3] = c[0] / (5.28 * Adu + K_blood * 1.0 / 40.0) + tsk  #  max flow = 90 [L/m2/h]/3600 <=> 1/40
				# Roots calculation #1
				if c[10] >= 0.0:  # Numerical safety to avoid negative roots
					tcore[5] = (-c[9] - math.pow(c[10], 0.5)) / (2.0 * c[4])
					tcore[0] = (-c[9] + math.pow(c[10], 0.5)) / (2.0 * c[4])
				# Roots calculation #2
				if c[8] >= 0.0:
					tcore[1] = (-c[6] + math.pow(abs(c[8]), 0.5)) / (2.0 * c[4])
					tcore[4] = (-c[6] - math.pow(abs(c[8]), 0.5)) / (2.0 * c[4])

				# Calculation of sweat losses  #
				tbody = 0.1 * tsk + 0.9 * tcore[j - 1]
				# Sweating flow calculation
				swm = 304.94 * (tbody - tbody_set) * Adu / 3600000.0
				# Saturation vapor pressure at temperature Tsk and for 100% HR
				vpts = 6.11 * math.pow(10.0, 7.45 * tsk / (235.0 + tsk))
				if tbody <= tbody_set:
					swm = 0.0
				if sex == 2:
					swm = 0.7 * swm
				esweat = -swm * Lvap
				hm = 0.633 * hc / (p * cair)  # Evaporation coefficient [W/(m^2*Pa)]
				fec = 1.0 / (1.0 + 0.92 * hc * rcl)
				emax = hm * (vpa - vpts) * Adu * Lvap * fec  # Max latent flux
				wetsk = esweat / emax  # skin wettedness
				# esw: Latent flux depending on w [W.m-2]
				if wetsk > 1.0:
					wetsk = 1.0
				eswdif = esweat - emax  # difference between sweating and max capacity
				if eswdif <= 0.0:
					esw = emax
				if eswdif > 0.0:
					esw = esweat
				if esw > 0.0:
					esw = 0.0
				ed = Lvap / (rdsk + rdcl) * Adu * (1.0 - wetsk) * (vpa - vpts)  # diffusion heat flux

				vb1 = tsk_set - tsk  # difference for the volume blood flow calculation
				vb2 = tcore[j - 1] - tc_set  #  idem
				if vb2 < 0.0:
					vb2 = 0.0
				if vb1 < 0.0:
					vb1 = 0.0
				# Calculation of the blood flow depending on the difference with the set value
				vb = (6.3 + 75 * vb2) / (1.0 + 0.5 * vb1)
				# energy balance MEMI modele
				enbal = h + ed + qresp + esw + csum + rsum
				# clothing temperature
				if count1 == 0:
					xx = 1.0
				if count1 == 1:
					xx = 0.1
				if count1 == 2:
					xx = 0.01
				if count1 == 3:
					xx = 0.001
				if enbal > 0.0:
					tcl = tcl + xx
				if enbal < 0.0:
					tcl = tcl - xx
				if (enbal > 0.0 or enbal2 <= 0.0) and (enbal < 0.0 or enbal2 >= 0.0):
					enbal2 = enbal
					count2 += 1
				else:
					break
			if count1 == 0.0 or count1 == 1.0 or count1 == 2.0:
				count1 = count1 + 1
				enbal2 = 0.0
			else:
				break
		# end "While True" (using 'break' statements)
		for k in range(20):
			g100 = 0
			if count1 == 3.0 and (j != 2 and j != 5):
				if j != 6 and j != 1:
					if j != 3:
						if j != 7:
							if j == 4:
								g100 = True
								break
						else:
							if tcore[j - 1] >= tc_set or tsk <= tsk_set:
								g100 = False
								break
							g100 = True
							break
					else:
						if tcore[j - 1] >= tc_set or tsk > tsk_set:
							g100 = False
							break
						g100 = True
						break
				else:
					if c[10] < 0.0 or (tcore[j - 1] < tc_set or tsk <= 33.85):
						g100 = False
						break
					g100 = True
					break
			if c[8] < 0.0 or (tcore[j - 1] < tc_set or tsk > tsk_set + 0.05):
				g100 = False
				break
		if g100 == False:
			continue
		else:
			if (j == 4 or vb < 91.0) and (j != 4 or vb >= 89.0):
				# Maximum blood flow
				if vb > 90.0:
					vb = 90.0
				# water loss in g/m2/h
				ws = swm * 3600.0 * 1000.0
				if ws > 2000.0:
					ws = 2000.0
				wd = ed / Lvap * 3600.0 * (-1000.0)
				wr = Eres / Lvap * 3600.0 * (-1000.0)
				wsum = ws + wr + wd
				return tcore[j - 1], tsk, tcl, esw
		# water loss
		ws = swm * 3600.0 * 1000.0  # sweating
		wd = ed / Lvap * 3600.0 * (-1000.0)  # diffusion = perspiration
		wr = Eres / Lvap * 3600.0 * (-1000.0)  # respiration latent
		wsum = ws + wr + wd
		if j - 3 < 0:
			index = 3
		else:
			index = j - 3
		return tcore[index], tsk, tcl, esw


def pet(tc, tsk, tcl, ta_init, esw_real):
	# Input variables of the PET reference situation:
	icl_ref = 0.9  # clo
	M_activity_ref = 80  # W
	v_air_ref = 0.1  # m/s
	vpa_ref = 12  # hPa
	icl = icl_ref

	tx = ta_init
	tbody = 0.1 * tsk + 0.9 * tc
	enbal2 = 0.0
	count1 = 0

	# base metabolism
	if sex == "male":
		met_base = 3.45 * math.pow(mbody, 0.75) * (1.0 + 0.004 * (30.0 - age) + 0.01 * (ht * 100.0 / math.pow(mbody, 1.0 / 3.0) - 43.4))
	else:
		met_base = 3.19 * math.pow(mbody, 0.75) * (1.0 + 0.004 * (30.0 - age) + 0.018 * (ht * 100.0 / math.pow(mbody, 1.0 / 3.0) - 42.1))
	# breathing flow rate
	rtv_ref = (M_activity_ref + met_base) * 1.44 * math.pow(10.0, -6.0)

	swm = 304.94 * (tbody - tbody_set) * Adu / 3600000.0  # sweating flow rate 
	vpts = 6.11 * math.pow(10.0, 7.45 * tsk / (235.0 + tsk))  # saturated vapour pressure at skin surface
	if tbody <= tbody_set:
		swm = 0.0
	if sex == "female":
		swm = swm * 0.7
	esweat = -swm * Lvap
	esweat = esw_real
	# standard environment
	hc = 2.67 + 6.5 * math.pow(v_air_ref, 0.67)
	hc = hc * math.pow(p / po, 0.55)
	# radiation saldo
	Aeffr = Adu * feff
	facl = (173.51 * icl - 2.36 - 100.76 * icl * icl + 19.28 * math.pow(icl, 3.0)) / 100.0
	if facl > 1.0:
		facl = 1.0
	# Increase of the exchange area depending on the clothing level
	fcl = 1 + (0.31 * icl)
	Acl = Adu * facl + Adu * (fcl - 1.0)
	hm = 0.633 * hc / (p * cair)  # Evaporation coefficient [W/(m^2*Pa)]
	fec = 1.0 / (1.0 + 0.92 * hc * 0.155 * icl_ref)  # vapour transfer efficiency for reference clothing
	emax = hm * (vpa_ref - vpts) * Adu * Lvap * fec  # max latetn flux for the reference vapour pressure 12 hPa
	wetsk = esweat / emax
	# skin wettedness
	if wetsk > 1.0:
		wetsk = 1.0
	eswdif = esweat - emax
	# diffusion
	ediff = Lvap / (rdsk + rdcl) * Adu * (1.0 - wetsk) * (vpa_ref - vpts)
	# esw: sweating [W.m-2] from the actual environment : in depends only on the difference with the core set temperature
	if eswdif <= 0.0:
		esw = emax
	if eswdif > 0.0:
		esw = esweat
	if esw > 0.0: 
		esw = 0.0

	while count1 != 4:
		rbare = Aeffr * (1.0 - facl) * emsk * sigm * (math.pow(tx + 273.2, 4.0) - math.pow(tsk + 273.2, 4.0))
		rclo = feff * Acl * emcl * sigm * (math.pow(tx + 273.2, 4.0) - math.pow(tcl + 273.2, 4.0))
		rsum = rbare + rclo  # Recalculation of the radiative losses
		# convection
		cbare = hc * (tx - tsk) * Adu * (1.0 - facl)
		cclo = hc * (tx - tcl) * Acl
		csum = cbare + cclo  # Recalculation of the convective losses
		# breathing
		texp = 0.47 * tx + 21.0
		Cres = cair * (tx - texp) * rtv_ref
		vpexp = 6.11 * math.pow(10.0, 7.45 * texp / (235.0 + texp))
		Eres = 0.623 * Lvap / p * (vpa_ref - vpexp) * rtv_ref
		qresp = (Cres + Eres)
		# ----------------------------------------
		# energy balance
		enbal = (M_activity_ref + met_base) + ediff + qresp + esw + csum + rsum
		if count1 == 0:
			xx = 1.0
		if count1 == 1:
			xx = 0.1
		if count1 == 2:
			xx = 0.01
		if count1 == 3:
			xx = 0.001
		if enbal > 0.0:
			tx = tx - xx
		if enbal < 0.0:
			tx += xx
		if (enbal > 0.0 or enbal2 <= 0.0) and (enbal < 0.0 or enbal2 >= 0.0):
			enbal2 = enbal
		else:
			count1 = count1 + 1
	return tsk, enbal, esw, ediff, tx, tcl

# real environment
# M_activity=80 # [W]
# icl=0.9
# Tair=21
# Tmrt=21
# v_air=0.1
# pvap=12. #Imposed value of Pvap

# tc, tsk, tcl, esw_real = system(Tair, Tmrt, pvap, v_air, M_activity, icl)
# tsk, enbal, esw, ed, PET, tcl = pet(tc, tsk, tcl, Tair, esw_real)

# print("PET value", round(PET,2))
# print("Tc ", round(tc,2)," / Tsk ", round(tsk,2)," / Tcl ", round(tcl,2))



'''
Created on 14 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from builtins import isinstance

from shapely.affinity import translate
from shapely.geometry import Point
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class Translation(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, vector):
        '''
        Constructor
        '''
        if not (isinstance(vector, Point) or isinstance(vector, (list, tuple))):
            raise IllegalArgumentTypeException(vector, 'Point, list or tuple') 
        self.xoff = vector.x if isinstance(vector, Point) else vector[0]
        self.yoff = vector.y if isinstance(vector, Point) else vector[1]
        if (isinstance(vector, Point) and vector.has_z):
            self.zoff = vector.z
        elif (isinstance(vector, (list, tuple)) and (3 == len(vector))):
            self.zoff = vector[2]
        else:
            self.zoff = None

    def runWithArgs(self, row):
        if self.zoff is None:
            return { 
                'geometry': translate(row.geometry, xoff=self.xoff, yoff=self.yoff) 
                }
        return { 
            'geometry': translate(row.geometry, xoff=self.xoff, yoff=self.yoff, zoff=self.zoff)
            }



'''
Created on 2 mai 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import cos, pi, sin, sqrt
from shapely.geometry import LineString, MultiLineString, Point
from shapely.ops import nearest_points
from shapely.prepared import prep
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class RayCasting3Lib(object):
    '''
    classdocs
    '''
    SQRT3DIV2 = sqrt(3) / 2
    DIVSQRT2 = 1 / sqrt(2)
    BASES = {
        12: [
            (1, 0), (SQRT3DIV2, 0.5), (0.5, SQRT3DIV2),
            (0, 1), (-0.5, SQRT3DIV2), (-SQRT3DIV2, 0.5),
            (-1, 0), (-SQRT3DIV2, -0.5), (-0.5, -SQRT3DIV2),
            (0, -1), (0.5, -SQRT3DIV2), (SQRT3DIV2, -0.5)
            ],
        8: [
            (1, 0), (DIVSQRT2, DIVSQRT2),
            (0, 1), (-DIVSQRT2, DIVSQRT2),
            (-1, 0), (-DIVSQRT2, -DIVSQRT2),
            (0, -1), (DIVSQRT2, -DIVSQRT2)
            ],
        6: [ (1, 0), (0.5, SQRT3DIV2), (-0.5, SQRT3DIV2), (-1, 0), (-0.5, -SQRT3DIV2), (0.5, -SQRT3DIV2) ],
        4: [ (1, 0), (0, 1), (-1, 0), (0, -1) ],
        3: [ (1, 0), (-0.5, SQRT3DIV2), (-0.5, -SQRT3DIV2) ],
        2: [ (1, 0), (-1, 0) ]
        }

    @staticmethod
    def areCovisibleIn2D(ptA, ptB, masks):
        # To avoid: "Inconsistent coordinate dimensionality"
        A = Point([ptA.x, ptA.y])
        B = Point([ptB.x, ptB.y])
        segm = LineString([A, B])

        psegm = prep(segm)
        hits = filter(psegm.intersects, masks.geometry)
        for maskGeom in hits:
            _geom = segm.intersection(maskGeom)
            _geom = _geom.geoms if GeomLib.isMultipart(_geom) else [_geom]
            for _g in _geom:
                if (not A.equals(_g)) and (not B.equals(_g)):
                    return False
        return True

    @staticmethod
    def areCovisibleIn3D(ptA, ptB, masks, maskElevationFieldname, masksSIdx=None):
        # if RayCasting3Lib.areCovisibleIn2D(ptA, ptB, masks):
        #     return True

        if masksSIdx is None:
            masksSIdx = masks.sindex

        # To avoid: "Inconsistent coordinate dimensionality"
        A = ptA if GeomLib.is3D(ptA) else GeomLib.forceZCoordinateToZ0(ptA, z0=0.0)
        B = ptB if GeomLib.is3D(ptB) else GeomLib.forceZCoordinateToZ0(ptB, z0=0.0)

        # SEGM [AB] MUST BE IN 2D (REGARDING INTERSECTION WITH 3D POLYGONS)
        segm2d = LineString([Point([A.x, A.y]), Point([B.x, B.y])])
        # alpha = arctan2(B.z - A.z, segm.length)
        alpha = (B.z - A.z) / segm2d.length

        # SELECTION OF A SUBSET OF MASKS
        collectionOfGeoms = []
        masksIds = masksSIdx.intersection(segm2d.bounds)
        for maskId in masksIds:
            row = masks.loc[maskId]
            collectionOfGeoms.append(
                GeomLib.forceZCoordinateToZ0(row.geometry, z0=row[maskElevationFieldname])
                )

        psegm = prep(segm2d)
        hits = filter(psegm.intersects, collectionOfGeoms)
        for maskGeom in hits:
            geom = segm2d.intersection(maskGeom)
            geom = geom.geoms if GeomLib.isMultipart(geom) else [geom]
            for g in geom:
                for c in g.coords:
                    p = Point(c)
                    if (not p.equals(A)):
                        beta = (p.z - A.z) / LineString([A, p]).length
                        if (beta >= alpha):
                            return False, segm2d
        return True, segm2d

    @staticmethod
    def __preparePanopticRays(n, i, angularOffset):
        for D in [12, 8, 6, 4, 3, 2]:
            if (0 == n % D):
                N = n // D
                if (0 == i % N):
                    # print(f'accel {D} pour {i}/{n}')
                    return RayCasting3Lib.BASES[D][i // N]
        return [float(cos(angularOffset * i)), float(sin(angularOffset * i))]

    @staticmethod
    def preparePanopticRays(nRays=64):
        angularOffset = (2.0 * pi) / nRays
        shootingDirs = [RayCasting3Lib.__preparePanopticRays(nRays, i, angularOffset) for i in range(nRays)]
        return shootingDirs

    @staticmethod
    def outdoorSingleRayCast2D(masks, viewPoint, shootingDir, rayLength=100.0):
        isAnInsidePoint = lambda point, polygon: ('0FFFFF212' == point.relate(polygon))

        # To avoid: "Inconsistent coordinate dimensionality"
        viewPoint = Point((viewPoint.x, viewPoint.y))
        remotePoint = Point((viewPoint.x + shootingDir[0] * rayLength, viewPoint.y + shootingDir[1] * rayLength))
        ray = LineString([viewPoint, remotePoint])

        hitPoint, hitDist = remotePoint, rayLength

        pray = prep(ray)
        hits = filter(pray.intersects, masks.geometry)
        for maskGeom in hits:
            # THE RAY STARTS FROM, HITS OR CROSSES THE BUILDING

            if (isAnInsidePoint(viewPoint, maskGeom)):
                # INDOOR VIEWPOINT
                return None, None, None

            _geom = ray.intersection(maskGeom)
            _geom = _geom.geoms if GeomLib.isMultipart(_geom) else [_geom]
            for _g in _geom:
                if (not viewPoint.equals(_g)) and (not _g.is_empty):
                    _, rp = nearest_points(viewPoint, _g)
                    _dist = viewPoint.distance(rp)
                    if (0.0 == _dist):
                        return LineString([viewPoint, viewPoint]), viewPoint, 0.0
                    elif (_dist < hitDist):
                        hitPoint, hitDist = rp, _dist

        return LineString([viewPoint, hitPoint]), hitPoint, hitDist

    @staticmethod
    def outdoorMultipleRayCast2D(masks, viewPoint, shootingDirs, rayLength=100.0):
        if not isinstance(masks, GeoDataFrame):
            raise IllegalArgumentTypeException(masks, 'GeoDataFrame')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')

        rays, hitPoints, hitDists = [], [], []
        for shootingDir in shootingDirs:
            ray, hitPoint, hitDist = RayCasting3Lib.outdoorSingleRayCast2D(
                masks, viewPoint, shootingDir, rayLength)
            if not ray is None:
                rays.append(ray)
                hitPoints.append(hitPoint)
                hitDists.append(hitDist)

        return MultiLineString(rays), hitPoints, hitDists



'''
Created on 8 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from geopandas.sindex import SpatialIndex
from numpy import arctan2, cos, pi, sin
from rtree.index import Index 
from shapely.geometry import LineString, MultiLineString, Point
from shapely.ops import nearest_points
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class RayCasting2Lib(object):
    '''
    classdocs
    '''

    @staticmethod
    def areCovisible(ptA, ptB, masksGdf, maskElevationFieldname, masksSIdx):
        # To avoid: "Inconsistent coordinate dimensionality"
        A = ptA if GeomLib.is3D(ptA) else GeomLib.forceZCoordinateToZ0(ptA, z0=0.0)
        B = ptB if GeomLib.is3D(ptB) else GeomLib.forceZCoordinateToZ0(ptB, z0=0.0)
        # segm = LineString([Point((ptA.x, ptA.y)), Point((ptB.x, ptB.y))])
        segm = LineString([A, B])
        alpha = arctan2(B.z - A.z, segm.length)

        masksIds = masksSIdx.intersection(segm.bounds)
        for maskId in masksIds:
            mask = masksGdf.loc[maskId]
            maskGeom = mask.geometry
            maskElevation = mask[maskElevationFieldname]

            if segm.intersects(maskGeom):
                tmpGeom = maskGeom.intersection(segm)
                gc = tmpGeom.geoms if GeomLib.isMultipart(tmpGeom) else [tmpGeom]
                for geom in gc:
                    for vertex in geom.coords:
                        hitPoint = Point(vertex)
                        _dist = A.distance(hitPoint)
                        _angle = arctan2(maskElevation - A.z, _dist)
                        if (_angle >= alpha):
                            return False, segm
        return True, segm

    @staticmethod
    def prepareOrientedRays(masksGdf, masksSIdx, viewPoint):
        _, nearestPoint, _ = GeomLib.getNearestFeature(
            masksGdf, masksSIdx, viewPoint, buffDist=40.0)
        u = GeomLib.unitVector([viewPoint.x, viewPoint.y], [nearestPoint.x, nearestPoint.y])
        return [u, [-u[0], -u[1]]]

    @staticmethod
    def preparePanopticRays(nRays=64):
        angularOffset = (2.0 * pi) / nRays
        shootingDirs = [[float(cos(angularOffset * i)), float(sin(angularOffset * i))] for i in range(nRays)]
        return shootingDirs

    @staticmethod
    def outdoorSingleRayCast2D(masksGdf, masksSIdx, viewPoint, shootingDir, rayLength=100.0): 
        # TL. 23.02.2021
        # To avoid: "Inconsistent coordinate dimensionality"
        viewPoint = Point((viewPoint.x, viewPoint.y))
        remotePoint = Point((viewPoint.x + shootingDir[0] * rayLength, viewPoint.y + shootingDir[1] * rayLength))
        ray = LineString([viewPoint, remotePoint])

        anchorId = GeomLib.getAnchoringBuildingId(viewPoint, masksGdf, masksSIdx)

        if anchorId is None:
            # ======================================================================
            # OUTDOOR VIEWPOINT
            hitPoint, hitDist, hitMask = None, rayLength, None

            masksIds = list(masksSIdx.intersection(ray.bounds))
            for maskId in masksIds:
                mask = masksGdf.loc[maskId]
                maskGeom = mask.geometry

                # if (ray.touches(maskGeom) or ray.crosses(maskGeom)):
                if (ray.crosses(maskGeom)):
                    # THE RAY HITS OR CROSSES THE BUILDING
                    _tmp = maskGeom.intersection(ray)
                    _, rp = nearest_points(viewPoint, _tmp)
                    dist = viewPoint.distance(rp)
                    if (dist < hitDist):
                        hitPoint, hitDist, hitMask = rp, dist, mask

            if hitPoint is not None:
                ray = LineString([viewPoint, hitPoint])
            else:
                hitPoint = remotePoint

        else:
            if GeomLib.isAnIndoorPoint(viewPoint, masksGdf, masksSIdx):
                # ======================================================================
                # INDOOR VIEWPOINT
                ray, hitPoint, hitDist, hitMask = None, None, 0.0, anchorId

            elif GeomLib.isABorderPoint(viewPoint, masksGdf, masksSIdx):
                # ======================================================================
                # VIEWPOINT ON A WALL
                buildingGeom = masksGdf.loc[anchorId].geometry
                _relate = ray.relate(buildingGeom)

                if (_relate in ['1FF00F212', 'F1FF0F212', 'F1FF0F212']):
                    # THE RAY GOES ALONG OR INTO THE "ANCHORING" BUILDING
                    ray, hitPoint, hitDist, hitMask = None, None, 0.0, anchorId

                elif ('FF1F00212' == _relate):
                    # THE RAY MOVES AWAY FROM THE "ANCHORING" BUILDING
                    hitPoint, hitDist, hitMask = None, rayLength, None

                    masksIds = list(masksSIdx.intersection(ray.bounds))
                    masksIds.remove(anchorId)
                    for maskId in masksIds:
                        mask = masksGdf.loc[maskId]
                        maskGeom = mask.geometry

                        if (ray.touches(maskGeom) or ray.crosses(maskGeom)):
                            # THE RAY HITS OR CROSSES THE "NON-ANCHORING" BUILDING
                            _tmp = maskGeom.intersection(ray)
                            _, rp = nearest_points(viewPoint, _tmp)
                            dist = viewPoint.distance(rp)
                            if (dist < hitDist):
                                hitPoint, hitDist, hitMask = rp, dist, mask

                    if hitPoint is not None:
                        ray = LineString([viewPoint, hitPoint])
                    else:
                        hitPoint = remotePoint

                else:
                    raise Exception('Unreachable instruction!')

            else:
                raise Exception('Unreachable instruction!')
        return [ray, hitPoint, hitDist, hitMask]

    @staticmethod
    def outdoorSingleRayCast25D(masksGdf, masksSIdx, viewPoint, shootingDir, rayLength,
                                elevationFieldName, background):
        # DEFAULT ALTITUDE OF THE VIEWPOINT
        h0 = viewPoint.z if (viewPoint.has_z) else 1.6

        # TL. 23.02.2021
        # To avoid: "Inconsistent coordinate dimensionality"
        viewPoint = Point((viewPoint.x, viewPoint.y))
        remotePoint = Point((viewPoint.x + shootingDir[0] * rayLength, viewPoint.y + shootingDir[1] * rayLength))
        ray = LineString([viewPoint, remotePoint])

        anchorId = GeomLib.getAnchoringBuildingId(viewPoint, masksGdf, masksSIdx)

        if anchorId is None:
            # ======================================================================
            # OUTDOOR VIEWPOINT
            hitPoint, hitDist, hitMask, hitHW = None, rayLength, None, 0.0

            masksIds = list(masksSIdx.intersection(ray.bounds))
            for maskId in masksIds:
                mask = masksGdf.loc[maskId]
                maskGeom = mask.geometry

                # if (ray.touches(maskGeom) or ray.crosses(maskGeom)):
                if (ray.crosses(maskGeom)):
                    # THE RAY HITS OR CROSSES THE BUILDING
                    _tmp = maskGeom.intersection(ray)
                    _, rp = nearest_points(viewPoint, _tmp)
                    height, dist = mask[elevationFieldName], viewPoint.distance(rp)
                    hw = (height - h0) / dist
                    if background and (hitHW < hw):
                        hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw
                    elif (not background) and (dist < hitDist):
                        hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw

            if hitPoint is not None:
                ray = LineString([viewPoint, hitPoint])
            else:
                hitPoint = remotePoint

        else:
            anchor = masksGdf.loc[anchorId]
            if GeomLib.isAnIndoorPoint(viewPoint, masksGdf, masksSIdx):
                # ======================================================================
                # INDOOR VIEWPOINT
                ray, hitPoint, hitDist, hitMask, hitHW = None, None, 0.0, anchor, None

            elif GeomLib.isABorderPoint(viewPoint, masksGdf, masksSIdx):
                # ======================================================================
                # VIEWPOINT ON A WALL
                buildingGeom = anchor.geometry
                _relate = ray.relate(buildingGeom)

                if (_relate in ['1FF00F212', 'F1FF0F212', 'F1FF0F212', 'F11F00212', '101F00212']):
                    # THE RAY GOES ALONG OR INTO THE "ANCHORING" BUILDING
                    ray, hitPoint, hitDist, hitMask, hitHW = None, None, 0.0, anchor, None

                elif ('FF1F00212' == _relate):
                    # THE RAY MOVES AWAY FROM THE "ANCHORING" BUILDING
                    hitPoint, hitDist, hitMask, hitHW = None, rayLength, None, 0.0

                    masksIds = list(masksSIdx.intersection(ray.bounds))
                    masksIds.remove(anchorId)
                    for maskId in masksIds:
                        mask = masksGdf.loc[maskId]
                        maskGeom = mask.geometry

                        if (ray.touches(maskGeom) or ray.crosses(maskGeom)):
                            # THE RAY HITS OR CROSSES THE "NON-ANCHORING" BUILDING
                            _tmp = maskGeom.intersection(ray)
                            _, rp = nearest_points(viewPoint, _tmp)
                            height, dist = mask[elevationFieldName], viewPoint.distance(rp)
                            hw = (height - h0) / dist
                            if background and (hitHW < hw):
                                hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw
                            elif (not background) and (dist < hitDist):
                                hitPoint, hitDist, hitMask, hitHW = rp, dist, mask, hw

                    if hitPoint is not None:
                        ray = LineString([viewPoint, hitPoint])
                    else:
                        hitPoint = remotePoint

                else:
                    raise Exception('Unreachable instruction!')

            else:
                raise Exception('Unreachable instruction!')
        return [ray, hitPoint, hitDist, hitMask, hitHW]

    @staticmethod
    def outdoorMultipleRayCast2D(masksGdf, masksSIdx, viewPoint, shootingDirs, rayLength=100.0):
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')

        rays, hitPoints, hitDists, hitMasks = [[], [], [], []]

        for shootingDir in shootingDirs:
            ray, hitPoint, hitDist, hitMask = RayCasting2Lib.outdoorSingleRayCast2D(
                masksGdf, masksSIdx, viewPoint, shootingDir, rayLength)
            if not ray is None:
                rays.append(ray)
                hitPoints.append(hitPoint)
            hitDists.append(hitDist)
            hitMasks.append(hitMask)

        return [MultiLineString(rays), hitPoints, hitDists, hitMasks]

    @staticmethod
    def outdoorMultipleRayCast25D(masksGdf, masksSIdx, viewPoint, shootingDirs, rayLength,
                                  elevationFieldName, background):
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        if not isinstance(masksSIdx, (SpatialIndex, Index)):
            raise IllegalArgumentTypeException(masksSIdx, 'SpatialIndex or Index')
        if not isinstance(viewPoint, Point):
            raise IllegalArgumentTypeException(viewPoint, 'Point')
        if elevationFieldName not in masksGdf:
            raise Exception('%s is not a relevant field name!' % (elevationFieldName))

        rays, hitPoints, hitDists, hitMasks, hitHWs = [[], [], [], [], []]

        for shootingDir in shootingDirs:
            ray, hitPoint, hitDist, hitMask, hitHW = RayCasting2Lib.outdoorSingleRayCast25D(
                masksGdf, masksSIdx, viewPoint, shootingDir, rayLength,
                elevationFieldName, background)
            if not ray is None:
                rays.append(ray)
                hitPoints.append(hitPoint)
            hitDists.append(hitDist)
            hitMasks.append(hitMask)
            hitHWs.append(hitHW)

        return [MultiLineString(rays), hitPoints, hitDists, hitMasks, hitHWs]



'''
Created on 26 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import timezone

from geopandas.geodataframe import GeoDataFrame
from numpy import isnan
from shapely.geometry import MultiPolygon, Polygon
from t4gpd.commons.DatetimeLib import DatetimeLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.ShadowLib import ShadowLib
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.sun.AbstractHardShadow import AbstractHardShadow


class STHardShadow(AbstractHardShadow):
    '''
    classdocs
    '''

    def __init__(self, occludersGdf, datetimes, occludersElevationFieldname='HAUTEUR',
                 altitudeOfShadowPlane=0, aggregate=False, tz=timezone.utc, model='pysolar'):
        '''
        Constructor
        '''
        if not isinstance(occludersGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(occludersGdf, 'GeoDataFrame')
        self.gdf = occludersGdf
        self.crs = occludersGdf.crs

        if occludersElevationFieldname not in occludersGdf:
            raise Exception('%s is not a relevant field name!' % (occludersElevationFieldname))
        self.occludersElevationFieldname = occludersElevationFieldname

        self.altitudeOfShadowPlane = altitudeOfShadowPlane
        self.aggregate = aggregate
        sunModel = SunLib(gdf=occludersGdf, model=model)

        self.sunPositions = DatetimeLib.fromDatetimesDictToListOfSunPositions(
            datetimes, sunModel, tz)

    def _auxiliary(self, row, radDir, solarAlti, solarAzim):
        occluderGeom = row.geometry
        occluderElevation = row[self.occludersElevationFieldname]

        if (occluderElevation is None) or (isnan(occluderElevation)):
            return occluderGeom

        if isinstance(occluderGeom, Polygon):
            _shadow = ShadowLib.projectBuildingOntoShadowPlane(
                occluderGeom, occluderElevation, radDir, self.altitudeOfShadowPlane)
            if _shadow is not None:
                return _shadow

        elif isinstance(occluderGeom, MultiPolygon):
            _shadows = []
            for g in occluderGeom.geoms:
                _shadow = ShadowLib.projectBuildingOntoShadowPlane(
                    g, occluderElevation, radDir, self.altitudeOfShadowPlane)
                if _shadow is not None:
                    _shadows.append(_shadow)
            return None if (0 == len(_shadows)) else MultiPolygon(_shadows)



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import abs, exp

from t4gpd.comfort.algo.CommonsLib import CommonsLib
from t4gpd.comfort.algo.ConstantsLib import ConstantsLib


class ETULib(object):
    '''
    classdocs
    '''

    @staticmethod
    def assess_etu(AirTC, RH, WS_ms, T_mrt, N, Albedo, SR01Up_1, SR01Dn_1, IR01UpCo_1, IR01DnCo_1):
        # Calculation of direct and diffuse shortwave radiation based on nebulosity approximation 
        # from MF weather station
        SDIF, Idn = CommonsLib.SDIF_Idn(N, SR01Up_1)

        # Air pressure in Pascal and mPa
        P_a, P_air = CommonsLib.airPressure(AirTC, RH)

        hc = CommonsLib.convectiveHeatTransferCoefficient(AirTC, WS_ms, P_a)

        # Fpcl permeation efficiency factor of clothing [-] after Huang (2006):
        Fpcl = 1 / (1 + 0.143 * hc * ConstantsLib.Icl)

        # frd effective radiation area factor [-] assumed to be alpha_eff 
        frd = ConstantsLib.alpha_eff

        # radiant heat transfer coefficient 
        hr = CommonsLib.radiantHeatTransferCoefficient(AirTC, RH, WS_ms, T_mrt)

        # combined heat transfer coefficient
        h = hc + hr

        # Fcl clothing thermal efficiency factor [-]
        Fcl = 1 / (1 + 0.155 * h * ConstantsLib.fcl * ConstantsLib.Icl)

        # Fcle effective clothing thermal efficiency [-]
        Fcle = Fcl * ConstantsLib.fcl

        hu = ((ConstantsLib.hco + ConstantsLib.hro)
               * ConstantsLib.Fcleo)  # 7.5 W.m-2.K-1
        # hv = (ConstantsLib.hco * ConstantsLib.Fcleo + 
        #       hr * Fcle)

        # "Lewis Relation" for air-vapor layer over skin surface
        La = (15.15 * (ConstantsLib.tsk + ConstantsLib.T_kel)
              / ConstantsLib.T_kel)
        # he 
        he = La * hc

        # he_prim_ETU = he * Fpcl * ConstantsLib.fcl

        # representative air temperature: the air temperature of an isothermal environment 
        # at predicted mean vote (PMV)=0
        T_ao = ConstantsLib.T_v

        #--------------------------------------------------
        # 1. TEMPERATURE
        #--------------------------------------------------
        # non-uniform air temperature field
        NUATF = (ConstantsLib.hco + ConstantsLib.hro) * (AirTC - T_ao)

        #--------------------------------------------------
        # 2. RADIATION
        #--------------------------------------------------
        # standard effective longwave radiant field
        SERFL = ConstantsLib.hro * ConstantsLib.Fcleo * (T_mrt - AirTC)

        # net direct solar radiation to the body [W.m-2]
        Rdir = (ConstantsLib.ah * ConstantsLib.fp
                * ConstantsLib.fcl * Idn)
        # net diffuse solar radiation to the body [W.m-2]
        Rdif = (ConstantsLib.ah * ConstantsLib.phi_skyh
                * frd * ConstantsLib.fcl * SDIF)
        # net reflected solar radiation to the body [W.m-2]
        Rref = (Albedo * ConstantsLib.ah 
                * ConstantsLib.phi_groundh * frd 
                * ConstantsLib.fcl * SR01Up_1)

        # effective shortwave radiant field (ERFS): effective temperature difference in air 
        # temperature caused by the heating power of solar radiation and is defined as the 
        # net solar energy exchanged on the exposed body surface (Rs or called short-wave 
        # solar radiation heat gain on human body) [W/m2]
        ERFS = Rdir + Rdif + Rref
        # Thermal velocity radiation-related Field (TVFr)
        TVFr = ((ConstantsLib.hro * ConstantsLib.Fcleo 
                 -hr * Fcle) * (ConstantsLib.tsk - T_mrt))

        #--------------------------------------------------
        # 3. Humidity
        #--------------------------------------------------
        # calculation of evaporative heat loss from sweating
        # mean body temperature [C]
        tb = (ConstantsLib.alpha * ConstantsLib.tsk + 
              (1 - ConstantsLib.alpha) * ConstantsLib.tcr)
        # mean neutral body temperature [C]
        tbn = (ConstantsLib.alpha * ConstantsLib.tskn + 
               (1 - ConstantsLib.alpha) * ConstantsLib.tcrn)
        Wsigb = abs(tb - tbn)
        Wsigsk = abs(ConstantsLib.tsk - ConstantsLib.tskn)
        # evaporative resistance of clothing [m2.kPa.W-1]
        Recl = ConstantsLib.Icl / (La * ConstantsLib.icl)
        # maximum evaporative heat loss
        Emax = (ConstantsLib.P_ssk - P_air) / (Recl + 1 / (ConstantsLib.fcl * he))
        # Emax2= he*(P_ssk-P_air) #Parsons(2003)
        # Mrsw rate at which sweat is secreted (...)
        Mrsw = 4.7 * 10 ** (-5) * Wsigb * exp(Wsigsk / 10.7)
        # evaporative heat loss from sweating
        Ersw = Mrsw * ConstantsLib.hfg
        # w skin wettedness [-] after Gagge et al.(1986) with values between 0.06 and 1 (dry-completely wet)
        w = 0.06 + 0.94 * (Ersw / Emax)

        # Thermal velocity evaporation-related field (TVFe)
        TVFe = ((w * ConstantsLib.heo * ConstantsLib.Fpclo 
                 * ConstantsLib.fclo - w * he * Fpcl 
                 * ConstantsLib.fcl) * (ConstantsLib.P_ssk - P_air))

        #--------------------------------------------------
        # 4. WIND
        #--------------------------------------------------
        # Thermal velocity field (TVF) [W/m2] is a measure of the cooling power of the actual air velocity and is defined as the net convective heat energy exchanged on the body surface.
        TVF = ((ConstantsLib.hco * ConstantsLib.Fcleo - 
                hc * Fcle) * (ConstantsLib.tsk - AirTC))

        # Effective velocity and clothing field 
        EVCF = TVF + TVFr + TVFe

        #===================================================
        # Calculation of ETU based on Watanabe et al. (2014)
        ETU = CommonsLib.iterationEtu(w, ConstantsLib.heo, P_air,
                                      T_ao, hu, NUATF, SERFL, ERFS, EVCF)

        # recalculate Hymidity with new ETU
        # P_ETUs saturated water vapor pressure at ETU, kPa
        P_ETUs = exp(18.956 - (4030.183 / (ETU + 235)))  # [mb]
        P_ETUs = P_ETUs / 10  # [kPa]

        # standard effective humid field
        SEHF = w * ConstantsLib.heo * (P_air - 0.5 * P_ETUs)

        # Effective temperature differences
        dif_NUATF = NUATF / hu
        dif_SERFL = SERFL / hu
        dif_ERFS = ERFS / hu
        dif_SEHF = SEHF / hu
        dif_EVCF = EVCF / hu

        return ETU, dif_NUATF, dif_SERFL, dif_ERFS, dif_SEHF, dif_EVCF



'''
Created on 28 mars 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import ceil, floor
from shapely.affinity import translate
from shapely.geometry import MultiPoint, MultiPolygon, Point, Polygon
from t4gpd.commons.GeomLib import GeomLib


class GridFaceLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def __findLongestEdge(geom):
        biggestLen, biggestEdge = 0, None
        for currEdge in GeomLib.toListOfBipointsAsLineStrings(geom):
            currLen = currEdge.length
            if (biggestLen < currLen):
                biggestLen, biggestEdge = currLen, currEdge
        return biggestEdge

    @staticmethod
    def __greatestEnclosingInt(value):
        return int(ceil(value) if (0 < value) else floor(value))

    @staticmethod
    def __preprocess(geom, du, dv):
        minx, miny, maxx, maxy = geom.bounds
        ll, lr, ur, ul = [minx, miny], [maxx, miny], [maxx, maxy], [minx, maxy]

        longestEdge = GridFaceLib.__findLongestEdge(geom)
        u = GeomLib.unitVector(*longestEdge.coords)
        v = [-u[1], u[0]]

        anchor = geom.exterior.coords[0]
        nuMin, nuMax, nvMin, nvMax = 0, 0, 0, 0

        for _p in (ll, lr, ur, ul):
            tmp = GeomLib.vector_to(anchor, _p)
            nu = GridFaceLib.__greatestEnclosingInt(GeomLib.dotProduct(u, tmp) / du)
            nv = GridFaceLib.__greatestEnclosingInt(GeomLib.dotProduct(v, tmp) / dv)
            if (nuMax < nu):
                nuMax = nu
            elif (nuMin > nu):
                nuMin = nu

            if (nvMax < nv):
                nvMax = nv
            elif (nvMin > nv):
                nvMin = nv

        return Point(anchor), u, v, nuMin, nuMax, nvMin, nvMax

    @staticmethod
    def __gridIntoFaces(geom, du, dv):
        if isinstance(geom, MultiPolygon):
            _mps = []
            for _geom in geom.geoms:
                _mp = GridFaceLib.__gridIntoFaces(_geom, du, dv)
                _mps += _mp.geoms
            return MultiPolygon(_mps)

        else:
            anchor, u, v, nuMin, nuMax, nvMin, nvMax = GridFaceLib.__preprocess(geom, du, dv)

            _geoms = []
            for nv in range(nvMin, nvMax + 1):
                for nu in range(nuMin, nuMax + 1):
                    p1 = translate(anchor,
                                   xoff=(nu - 0.5) * du * u[0] + (nv - 0.5) * dv * v[0],
                                   yoff=(nu - 0.5) * du * u[1] + (nv - 0.5) * dv * v[1])
                    p2 = translate(anchor,
                                   xoff=(nu + 0.5) * du * u[0] + (nv - 0.5) * dv * v[0],
                                   yoff=(nu + 0.5) * du * u[1] + (nv - 0.5) * dv * v[1])
                    p3 = translate(anchor,
                                   xoff=(nu + 0.5) * du * u[0] + (nv + 0.5) * dv * v[0],
                                   yoff=(nu + 0.5) * du * u[1] + (nv + 0.5) * dv * v[1])
                    p4 = translate(anchor,
                                   xoff=(nu - 0.5) * du * u[0] + (nv + 0.5) * dv * v[0],
                                   yoff=(nu - 0.5) * du * u[1] + (nv + 0.5) * dv * v[1])

                    _cell = Polygon([p1, p2, p3, p4])
                    if geom.contains(_cell):
                        _geoms.append(_cell)

            return MultiPolygon(_geoms)

    @staticmethod
    def grid(geom, dx, dy=None, intoPoint=True):
        if not GeomLib.isPolygonal(geom):
            return { 'geometry': MultiPolygon(), 'n_cells': 0 }

        dy = dx if dy is None else dy
        geom = GeomLib.normalizeRingOrientation(geom)
        geom = GridFaceLib.__gridIntoFaces(geom, dx, dy)

        if intoPoint:
            geom = MultiPoint([_p.centroid for _p in geom.geoms])

        return { 'geometry': geom, 'n_cells': len(geom.geoms) }



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan
from pandas.core.frame import DataFrame
from t4gpd.comfort.algo.ETULib import ETULib
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class ETU(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg', WS_ms='WS_ms_Avg',
                 T_mrt='T_mrt', N='N', Albedo='Albedo_1_Avg', SRUp='SR01Up_1_Avg',
                 SRDn='SR01Dn_1_Avg', IRUp='IR01UpCo_1_Avg', IRDn='IR01DnCo_1_Avg'):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]
        T_mrt: Mean radiant temperature [C]
        N:
        Albedo:
        SRUp:
        SRDn:
        IRUp:
        IRDn:
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (AirTC, RH, WS_ms, T_mrt, N, Albedo, SRUp, SRDn,
                          IRUp, IRDn):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.AirTC = AirTC
        self.RH = RH
        self.WS_ms = WS_ms
        self.T_mrt = T_mrt

        self.N = N
        self.Albedo = Albedo
        self.SRUp, self.SRDn = SRUp, SRDn
        self.IRUp, self.IRDn = IRUp, IRDn

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        RH = row[self.RH]
        WS_ms = row[self.WS_ms]
        T_mrt = row[self.T_mrt]

        N = row[self.N]
        Albedo = row[self.Albedo]
        SRUp, SRDn = row[self.SRUp], row[self.SRDn]
        IRUp, IRDn = row[self.IRUp], row[self.IRDn]

        if (N is None) or isnan(N):
            return {
                'ETU': None,
                'dif_NUATF': None,
                'dif_SERFL': None,
                'dif_ERFS': None,
                'dif_SEHF': None,
                'dif_EVCF': None
                }

        ETU, dif_NUATF, dif_SERFL, dif_ERFS, dif_SEHF, dif_EVCF = ETULib.assess_etu(
            AirTC, RH, WS_ms, T_mrt, N, Albedo, SRUp, SRDn, IRUp, IRDn)

        return {
            'ETU': ETU,
            'dif_NUATF': dif_NUATF,
            'dif_SERFL': dif_SERFL,
            'dif_ERFS': dif_ERFS,
            'dif_SEHF': dif_SEHF,
            'dif_EVCF': dif_EVCF
            }



'''
Created on 23 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from shapely.affinity import scale 
from t4gpd.commons.GeoProcess import GeoProcess

from t4gpd.pyvista.commons.Triangle3D import Triangle3D


class GeodeCiel(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, norecursions, radius=None):
        '''
        Constructor
        '''
        self.listOfTriangles = [
            Triangle3D([0, 1, 0], [0, 0, 1], [1, 0, 0]),
            Triangle3D([0, 1, 0], [0, 0, -1], [1, 0, 0])
            ]
        self.norecursions = norecursions
        self.radius = radius

    def __divide(self, listOfTriangles, norecursions):
        result = []
        if (0 >= norecursions):
            for tri in listOfTriangles:
                a, b, c = tri.ptA, tri.ptB, tri.ptC
                # Add NE triangle
                result.append(tri)
                # Add NW triangle
                result.append(Triangle3D(self.__mirrorX(a), self.__mirrorX(b), self.__mirrorX(c)))
                # Add SE triangle
                result.append(Triangle3D(self.__mirrorY(a), self.__mirrorY(b), self.__mirrorY(c)))
                # Add SW triangle
                result.append(Triangle3D(self.__mirrorXY(a), self.__mirrorXY(b), self.__mirrorXY(c)))
        else:
            for tri in listOfTriangles:
                result += self.__divide(tri.divideBy4OnSphere(), norecursions - 1)
        return result

    def __mirrorX(self, pt):
        return [ -pt[0], pt[1], pt[2] ]

    def __mirrorY(self, pt):
        return [ pt[0], -pt[1], pt[2] ]

    def __mirrorXY(self, pt):
        return [ -pt[0], -pt[1], pt[2] ]

    @staticmethod
    def fromNRaysToNRecursions(nrays):
        return 0 if (8 >= nrays) else 1 + GeodeCiel.fromNRaysToNRecursions(nrays / 4)

    @staticmethod
    def getNRaysOptions():
        # return (4, 16, 64, 256, 1024, 4096, 16384, 65536, 262144)
        return (8, 32, 128, 512, 2048, 8192, 32768, 131072, 524288)

    def getRays(self):
        rows = self.__divide(self.listOfTriangles, self.norecursions)
        return [row.centroid() for row in rows]

    def run(self):
        rows = self.__divide(self.listOfTriangles, self.norecursions)
        if (self.radius is None) or (1 == self.radius):
            rows = [{'geometry': row.toPolygon()} for row in rows]
        else:
            rows = [{'geometry': scale(
                row.toPolygon(),
                xfact=self.radius,
                yfact=self.radius,
                zfact=self.radius,
                origin=(0, 0, 0))} for row in rows]
        return GeoDataFrame(rows)
